from builtin.math import *
from builtin.view import *

# > sort 7
ENGINE_NAME = "ATEngine"


class Scheduler:  # > remove
    now: int = 0  # > remove

    def __init__(self):  # > remove
        self.updates: list = None  # > remove

    def handle(self):  # > remove
        pass  # > remove

    def add(self, obj, *args) -> "CreepLogic" | None:  # > remove
        pass  # > remove

    def find(self, name: str) -> object | list[object] | None:  # > remove
        pass  # > remove


class StageMachineLogicMeta(type):
    """
    合并后的元类，用于自动注册类到Scheduler中，并管理Stage类的状态节点和原型记录。

    功能：
    - 如果类名不是'CreepLogic'，则会通过Scheduler的login方法将其注册。
    - 自动发现并记录状态节点（函数），同时将类原型记录到__protos__中。
    """

    logs = []

    __recursive__ = []
    """
    记录最近一个类定义时使用的装饰器。
    * 创建完一个类后，会清空这个字典。
    * recursive装饰器会将当前类名记录在这里。
    """

    __protos__ = {}
    """
    字典，用于记录Stage类的原型，键为类名，值为类对象。
    """

    IGNORES = [
        'handle', 'loading', 'changed', "ref", 'handle0', 'leader', "mates", 'creeps',
        "removed", "killed", "birth", "step0", "step", "step2", "draw", "team",
        "name", "py_name", "listenFor", "cancelListen", "pushEvent", "initialize",
        "link", "relink", "refLast", "refNext", "refAdjacent", "enable", "disable", "remove", "transform",
        "productQueue", "showQueue", "jumpQueue", "schedule", "cancelSchedule",
    ]
    """
    列表，包含一些特殊方法名，这些方法不会被识别为状态节点。
    """

    def __new__(meta, name, bases, attrs):
        """
        创建新类并自动注册到Scheduler中（若类名不是'CreepLogic'），同时管理状态节点和原型记录。

        Args:
            meta (type): 元类本身，即StageMachineLogicMeta。
            name (str): 新创建的类的名称。
            bases (tuple): 新创建的类的基类元组。
            attrs (dict): 新创建的类的属性字典。

        Returns:
            type: 新创建的类。
        """

        # 在attrs中添加一个__states__属性，用于记录状态节点
        attrs['__states__'] = []
        # 在attrs中添加一个__recursive__属性，用于记录需要连续执行的状态节点
        attrs['__recursive__'] = StageMachineLogicMeta.__recursive__
        # 清空__recursive__，避免影响其他类
        StageMachineLogicMeta.__recursive__ = []

        # 将每个父类的__states__和__recursive__属性合并到attrs中
        for base in bases:
            # StageMachineLogicMeta.logs.append(base.NAME)

            # StageMachineLogicMeta.logs.append(base.__states__)
            # StageMachineLogicMeta.logs.append(base.__recursive__)
            if base.__states__:  # js中运行才能这么写
                attrs['__states__'].extend(base.__states__)
            if base.__recursive__:
                attrs['__recursive__'].extend(base.__recursive__)

        # 去重
        attrs['__states__'] = list(set(attrs['__states__']))
        attrs['__recursive__'] = list(set(attrs['__recursive__']))

        # 遍历属性，找出状态节点（函数）
        # > insert const iters = Object.entries(attrs);
        iters = []  # > remove
        for key, value in iters:
            # 假设不以'_'开头且是函数，记录到__states__中
            if not key.startswith('_') and callable(value):
                if key.isupper():
                    continue
                if key in StageMachineLogicMeta.IGNORES:
                    continue
                attrs['__states__'].append(key)

        # StageMachineLogicMeta.logs.append(f"{name} {attrs['NAME']:>20} {len(attrs['__states__'])} {attrs['__recursive__']}")

        # 还原__new__方法为默认的，避免修改基类
        if name != 'Stage':
            attrs['__new__'] = object.__new__

        # 创建类
        new_class = type.__new__(meta, name, bases, attrs)

        # 如果类名不是'CreepLogic'，则将其注册到Scheduler中
        if name != "CreepLogic":
            # 将新创建的类添加到全局的Scheduler中
            # 此处假设Scheduler是一个全局可用的类，并且具有login方法用于注册类
            Scheduler().login(new_class)

        # 记录到__protos__中，方便后续获取
        if name != 'Stage':
            StageMachineLogicMeta.__protos__[new_class.NAME] = new_class

        return new_class

    @classmethod
    def proto(cls, name):
        """
        获取一个Stage类的原型。

        Args:
            name (str): 类名。

        Returns:
            type: Stage类的原型。

        Raises:
            ValueError: 如果类名不存在于__protos__中。
        """
        if name in cls.__protos__:
            return cls.__protos__[name]
        else:
            # raise ValueError(f"Stage proto '{name}' not found")
            if LANGUAGE == 'cn':
                std.error("StageMachineLogicMeta.proto", "找不到Stage原型: " + name)
            else:
                std.error("StageMachineLogicMeta.proto", "Stage prototype not found: " + name)


def recursive(function_name):
    """
    递归装饰器，用于标记一个函数为递归调用。
    * 被递归装饰的函数，会在一个tick内被多次调用，直到返回一个新的状态。
    * 注意，最大递归深度为16。通过Stage的子类的LIMIT属性可以修改。
    """
    # 编译器将上面一句话的function_name替换为对应的函数名
    print("__recursive__", StageMachineLogicMeta.__recursive__)
    StageMachineLogicMeta.__recursive__.append(function_name)

    def _inner_(f):
        return f

    return _inner_


class SitePlan:
    def __init__(self, ref: Point, dx: int, dy: int, building_type: type):
        """
        初始化一个建筑规划点

        Args:
            ref (Point): 参考点
            dx (int): 相对于参考点的x轴偏移量
            dy (int): 相对于参考点的y轴偏移量
            building_type (type): 建筑类型
        """
        self.ref = ref
        self.offset = Point(dx, dy)
        self.building_type = building_type
        self.site = None

    def prepare(self) -> int:
        """
        准备建筑点

        计算实际建筑点位置，并尝试创建建筑点

        Returns:
            int: 返回创建建筑点的结果标识
        """
        self.point = self.ref.offset(self.offset.x, self.offset.y)
        self.site = ConstructionSite.plan(self.point.x, self.point.y, self.building_type)
        return self.site

    def __str__(self):
        return f"SitePlan({self.point}, {self.building_type})"


class SitePlaner:
    def __init__(self, plans: list[SitePlan]):
        """
        初始化一个建筑规划器

        Args:
            plans (list[SitePlan]): 建筑规划点列表
        """
        self.plans = plans
        self.site: ConstructionSite = None
        self.cur_idx = 0
        self.max_idx = len(plans) - 1
        self._current_prepare: SitePlan = None
        self._last_built = None  # 最近的完成的建筑

    @property
    def finished(self) -> bool:
        """
        判断当前建筑点是否已经建造完毕

        Returns:
            bool: 如果当前建筑点已经建造完毕则返回True，否则返回False
        """
        return self.cur_idx >= self.max_idx and not st.site(self.site)

    @property
    def structure(self) -> Structure | None:
        """
        获取最近完成的建筑
        """
        return self._last_built

    def next(self) -> ConstructionSite | None:
        """
        获取下一个建筑点

        如果当前建筑点已经建造完毕或不存在但是有储备建造点，则弹出并返回下一个建筑点

        Returns:
            SitePlan | None: 下一个建筑点，如果没有则返回None
        """
        if self.cur_idx > 0:  # 更新上一次完工的建筑
            if self.site and not st.site(self.site) and self.site.structure:
                self._last_built = self.site.structure

        # 检查最近完工的建筑的存活性
        if self._last_built and not self._last_built.exists:
            self._last_built = None

        if self._current_prepare is None and self.cur_idx <= self.max_idx:
            tmp = self.plans[self.cur_idx]
            self.cur_idx += 1
            _pr = tmp.prepare()
            if st.jnumber(_pr):
                print(f"准备建筑点失败: {tmp} -> {_pr}")
                return self.next()
            self._current_prepare = tmp

        if not st.site(self.site) or not self.site.exists or self.site.progress >= self.site.progressTotal:
            self.site = None

        if not self.site and self._current_prepare is not None:
            self.site = self._current_prepare.site
            self._current_prepare = None

        return self.site

    def remove(self):
        """
        移除当前建筑点

        如果存在当前建筑点，则移除它，并清空相关状态
        """
        if st.site(self.site):
            self.site.remove()
        self.site = None

        if self._current_prepare is not None:
            self._current_prepare.site.remove()
            self._current_prepare = None

    def __str__(self):
        return f"SitePlaner({self.plans})"


class _CreepLogicSettings:
    """
    爬虫逻辑设置类，用于定义爬虫的基本属性和行为。

    Attributes:
        NAME (str): 爬虫名称。
        LINK (list[str]): 爬虫逻辑链接，标记的link只会启用一次，即使反复初始化。
        ONCE (bool): 如果killed函数被调用，置为True，爬虫逻辑将被删除。
        RECIPE (list[str]): 爬虫配方，默认为[MOVE]。
        OPTIMISE (bool): 是否自动优化配方，默认为True。
        EXTENSION (bool): 生产爬虫时是否使用extension，默认为True。
    """

    # ---------------------- 爬虫本体 | Creep --------------------------- #
    NAME: str = "creep"  # 爬虫名称
    DRAW: bool = False  # 是否绘制爬虫信息
    LAYER: int = 10  # 爬虫绘制图层，默认为10

    # ---------------------- 流程与结构 | Flow & Struct --------------------------- #
    LOG: bool = True  # 是否输出状态机日志
    LINK: list[str] = None  # 爬虫逻辑链接
    ONCE: bool = False  # 爬虫逻辑删除标志
    LIMIT: int = 16  # 递归深度限制
    STAGE: int = 0  # 爬虫执行逻辑的阶段(0(阶段0前), 1(阶段1前), 2(阶段2前), 3(阶段3前))

    # ---------------------- 孵化选项 | Spawning --------------------------- #
    VIRTUAL: bool = False  # 是否虚拟逻辑(每次handle都会自动将SPAWNABLE置为False)
    SPAWNABLE = True  # 是否可以孵化标志(如果没有孵化标志，将不会被孵化; 没有孵化的逻辑不会执行后续step除非拥有虚拟逻辑标志)
    RECIPE: list[str] | NamedRecipe = ["MOVE"]  # 爬虫配方，默认为["MOVE"]
    OPTIMISE: bool = True  # 自动优化配方标志
    EXTENSION: bool = True  # 使用extension标志
    PRIORITY: int = 0  # 孵化优先级
    DIRECTION: int = None  # 出生方向


class _CreepLogicBase(_CreepLogicSettings):
    """
    爬虫逻辑基类，继承自_CreepLogicSettings，提供爬虫逻辑的基本属性和方法。

    Attributes:
        __kind__ (str): 爬虫逻辑的种类标识。
        __id__ (dict): 用于记录爬虫逻辑的ID。
        __names__ (dict): 用于记录爬虫逻辑的名称。
        __types__ (dict): 用于记录爬虫逻辑的类型。
    """

    __kind__ = "creep-logic"
    __id__ = {}  # 用于记录爬虫逻辑的ID
    __names__ = {}  # 用于记录爬虫逻辑的名称
    __types__ = {}  # 用于记录爬虫逻辑的类型
    __states__ = []  # 这个属性由元类自动添加，除此外只读

    def _after_initial(self, sc: StructureSpawn | Creep):
        # 根据传入的sc对象类型，分别赋值给creep或spawn
        if st.spawn(sc):
            self._spawn = sc
        elif st.creep(sc):
            self._creep = sc
            self._creep.logic = self
            self._creep.name = self._name_
            # > insert self._creep.name = self._name_;

        # 检查recipe是否为空，为空则报错
        if not self._recipe:
            std.error("CreepLogic", "CreepLogic must has recipe.")

        # 对name进行一系列检查
        name_len = len(self._name)
        if name_len <= 0:
            std.error("CreepLogic", "CreepLogic must has not empty name.")
        if self._name[name_len - 1].isdigit():
            self._name += "_"
        if self._name in CreepLogic.__types__ and self.__class__.__name__ is not CreepLogic.__types__[self._name]:
            std.error("CreepLogic", "CreepLogic must has unique type: " + self.__class__.__name__)
        else:
            CreepLogic.__types__[self._name] = self.__class__.__name__
        if self._name not in CreepLogic.__names__:
            CreepLogic.__names__[self._name] = {}
        CreepLogic.__names__[self._name][self._cid] = self

        # 处理link相关逻辑
        if self.__class__.LINK:
            self.link(*self.__class__.LINK)
            self.__class__.LINK = None

    @staticmethod
    def creeps(*logics) -> list[Creep | None]:
        """
        * 将多个爬虫逻辑实例转换为爬虫实例列表 | Convert multiple CreepLogic instances to a list of Creep instances
        如果逻辑没有绑定creep，则对应位置返回None | If the logic is not bound to a creep, return None at the corresponding position
        """
        # 将传入的爬虫逻辑实例转换为爬虫实例列表
        return [logic._creep if st.creep(logic._creep) else None for logic in logics]

    @classmethod
    def _get_id_(cls, _name):
        """
        获取指定名称的唯一ID，并自增该名称的ID计数。

        Args:
            _name (str): 名称。

        Returns:
            int: 唯一ID。
        """
        # 如果名称不在__id__字典中，则初始化其ID为-1
        if _name not in cls.__id__:
            cls.__id__[_name] = -1
        # 自增名称的ID计数
        cls.__id__[_name] += 1
        # 返回自增后的ID
        return cls.__id__[_name]

    @classmethod
    def _seek_id(cls, _name):
        """
        查询指定名称的当前ID。

        Args:
            _name (str): 名称。

        Returns:
            int: 当前ID，若名称不存在则返回-1。
        """
        # 如果名称不在__id__字典中，则返回-1
        if _name not in cls.__id__:
            return -1
        # 返回名称的当前ID
        return cls.__id__[_name]

    @classmethod
    def ref(cls, name: str) -> list | "CreepLogic" | None:
        """
        * 通过名称获取CreepLogic实例 | Get CreepLogic instance by name
        如果你的name不以数字结尾，将返回一个列表，否则返回一个CreepLogic实例 |
            If your name does not end with a number, a list will be returned, otherwise a CreepLogic instance will be returned
        Args:
            name: str  # CreepLogic名称 | CreepLogic name

        Returns:
            list | CreepLogic  # CreepLogic实例或列表 | CreepLogic instance or list

        """
        if st.na(name): return None
        name_len = len(name)
        if name_len <= 0:
            std.error("CreepLogic.ref", "CreepLogic name must not be empty.")

        if not name[name_len - 1].isdigit():
            insts = cls.__names__.py_get(name, {})
            insts = list(insts.values())
            return insts
        else:
            if name_len <= 1:
                std.error("CreepLogic.ref", "CreepLogic name can not be only number: " + name)
            # reversed search a not digit
            for i in range(name_len - 2, -1, -1):
                if not name[i].isdigit():
                    name, num = name[:i + 1], int(name[i + 1:])
                    insts = cls.__names__.py_get(name, {})
                    insts = list(insts.values())
                    if num < len(insts):
                        return insts[num]
                    else:
                        return None
            return None

    def initialize(self, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        当__init__函数执行完毕时立刻调用此 | called when the __init__ function is executed
        Args:
            k: GlobalKnowledge

        Returns:
            not care
        """
        pass

    def loading(self, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        当首次以enabled状态进入handle函数时调用此 | called when first entering the handle function with enabled state

        Args:
            k: GlobalKnowledge

        Returns:
            not care
        """
        pass

    def removed(self, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        * 由外部停止此对象时调用此 | called when this object is stopped externally
        Args:
            k: GlobalKnowledge

        Returns:
            None
        """
        pass

    def killed(self, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        * 当绑定的Creep被击杀时调用此 | called when the bound Creep is killed

        Args:
            k: GlobalKnowledge

        Returns:
            not care
        """
        if self._once:
            self.remove()

    def birth(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        当自身对应RECIPE的Creep被创建完成时调用此 | called when the Creep corresponding to RECIPE is created

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def step0(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        stage: StepBefore
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def step(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        stage: Step
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def step2(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        stage: StepAfter
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def draw(self, c: Creep, v: View, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        stage: Draw
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep
            v: View
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        v.header(c)

    # ------------------------- StageMachine ------------------------------------

    def changed(self, src: str, dst: str, k: GlobalKnowledge):
        """
        在不同状态切换时调用，子类可重写。

        Args:
            src (str): 源状态。
            dst (str): 目标状态。
            k (GlobalKnowledge): 全局知识对象。
        """
        pass

    def handle0(self, log=True):
        """
        执行当前状态，若返回一个str，则切换到该状态。

        Args:
            log (bool): 是否打印日志，默认为True。

        Raises:
            ValueError: 如果当前状态或目标状态无效。
        """
        if self.current not in self.__states__:
            # raise ValueError(f"Invalid state '{self.current}' in Stage '{self.NAME}'")
            if LANGUAGE == 'cn':
                std.error(self.name + ".stageMachine", "无效的状态: " + self.current)
            else:
                std.error(self.name + ".stageMachine", "Invalid state: " + self.current)

        try:
            next_state = getattr(self, self.current)(self.creep, know, *self._children)
        except Exception as e:
            if LANGUAGE == 'cn':
                std.error(self.name + "." + self.current, "执行状态逻辑时发生错误: " + str(e))
            else:
                std.error(self.name + "." + self.current,
                          "An error occurred while executing the state logic: " + str(e))
            return
        if not isinstance(next_state, str):
            next_state = self.current
        if next_state not in self.__states__:
            # raise ValueError(f"Invalid state '{next_state}' in Stage '{self.NAME}'")
            if LANGUAGE == 'cn':
                std.error(self.name + ".stageMachine", "无效的新状态: " + next_state)
            else:
                std.error(self.name + ".stageMachine", "Invalid next state: " + next_state)

        # 检查是否recursive
        recursive_flag = False
        if (self.current in self.__recursive__) or (next_state in self.__recursive__):
            self._recursive_left -= 1
            if self._recursive_left <= 0:
                if LANGUAGE == 'cn':
                    std.error(self.name + ".stageMachine", "递归调用次数超过限制: " + str(self.LIMIT))
                else:
                    std.error(self.name + ".stageMachine", "Recursive call count exceeded limit: " + str(self.LIMIT))
                return
            recursive_flag = True

        if log:
            std.info(self.name + ".stageMachine", self.current + " -> " + next_state, "..." if recursive_flag else "")

        try:
            if self.current != next_state: self.changed(self.current, next_state, know)
        except Exception as e:
            if LANGUAGE == 'cn':
                std.error(self.name + ".changed", "执行状态切换逻辑时发生错误: " + str(e))
            else:
                std.error(self.name + ".changed", "An error occurred while executing the state change logic: " + str(e))
        self.current = next_state

        if recursive_flag: self.handle0(log)


class CreepLogic(_CreepLogicBase, metaclass=StageMachineLogicMeta):
    def __init__(self, spawn_creep: StructureSpawn | Creep = None):
        """
        初始化爬虫逻辑 | Initialize creep logic

        Args:
            spawn_creep: StructureSpawn | Creep | None # 用于初始化的spawn或creep | spawn or creep for initialization
        """
        # 初始化基本属性
        self._team: TeamLogic | None = None
        # 当改逻辑被绑定到队伍逻辑时，指向队伍逻辑 | When this logic is bound to the team logic, point to the team logic
        self._parent: CreepLogic | None = None
        # 当由其他爬虫逻辑创建时，指向父逻辑 | When created by other creep logic, point to the parent logic
        self._auto: bool = False
        # 是否在stage=0的handle中自动置为enable | Whether to automatically set enable in handle of stage=0
        self._creep: Creep | None = None
        # 绑定的creep | bound creep
        self._spawn: StructureSpawn | None = None
        # 绑定的spawn | bound spawn
        self._alive: bool = True
        # 是否存活. 一旦被置位为False，将不再被更新 | Whether alive. Once set to False, it will no longer be updated
        self._inited: bool = False
        # 是否已经初始化 | Whether initialized
        self._loaded: bool = False
        # 是否已经加载 | Whether loaded
        self._enabled: bool = True
        # 被置位时，才允许执行更新逻辑(但不影响创建creep) | When set, only allow execution of update logic (but does not affect creating creep)
        self._scheduler: Scheduler = Scheduler()
        self._next_links: list = []
        self._children: list = []
        # ----------------------
        self._log: bool = self.LOG              # 是否输出日志
        self._name: str = self.NAME             # 爬虫类名
        self._draw: bool = self.DRAW            # 是否绘制爬虫信息(在creep的头上)
        self._once: bool = self.ONCE            # 是否一次性
        self._priority: int = self.PRIORITY     # 孵化优先级
        self._recipe: list[str] = self.RECIPE   # 配方
        self._optimise: bool = self.OPTIMISE    # 是否自动优化配方
        self._virtual: bool = self.VIRTUAL      # 是否为虚拟爬虫(不创建creep也可执行逻辑, c为None)
        self._spawnable: bool = self.SPAWNABLE  # 是否可以孵化(否则视作虚拟逻辑)
        self._extension: bool = self.EXTENSION  # 是否使用extension中的能量
        self._direction: int = self.DIRECTION   # 出生方向
        self._stagelv: int = self.STAGE         # 状态机阶段
        # --------------------------------
        self._recipe = self._recipe.recipe if self._recipe.recipe else self._recipe
        self._cid: int = self._get_id_(self._name)  # creep-id
        self._view: View | None = View(self.LAYER) if self._draw else None
        self._cost: int = PartsVector.partsCost(self._recipe)
        self.name: str = self._name + str(self._cid)
        self._expand_name: str = self._expand_name_(self.name, 16)  # 用于输出的爬虫名
        # > self.name = self.name_;
        self._eda = None  # > remove
        # > insert self._eda = new ListenSystem(self, self.py_name);
        self._scs = None  # > remove
        # > insert self._scs = new ScheduleSystem(self);

        # stageMachine
        self._recursive_left = self.LIMIT
        self.current: str = self.__states__[0] if len(self.__states__) else ""

        # 初始化&检查
        self._after_initial(spawn_creep)

        # initialize
        self.initialize(know, *self._children)

    @property
    def x(self) -> int | undefined:
        """
        获取creep的x坐标
        * 如果creep不存在，则返回undefined
        """
        if not self._creep or not self._creep.exists:
            return undefined
        return self._creep.x

    @property
    def y(self) -> int | undefined:
        """
        获取creep的y坐标
        * 如果creep不存在，则返回undefined
        """
        if not self._creep or not self._creep.exists:
            return undefined
        return self._creep.y

    @property
    def id(self) -> str:
        """
        获取creep的id
        * 如果creep未绑定，则返回空字符串
        """
        if not self._creep:
            return ""
        return self._creep.id

    @property
    def exists(self) -> bool:
        """
        获取creep是否存在
        * 如果creep未绑定，则返回False
        """
        if not self._creep:
            return False
        return self._creep.exists

    @property
    def cid(self) -> int:
        """
        获取creep的id  (creep-id)
        """
        return self._cid

    @property
    def protoname(self) -> str:
        """
        获取creeplogic的默认类名
        """
        return self._name

    @property
    def cost(self) -> int:
        """
        获取recipe的cost
        """
        return self._cost

    @property
    def grade(self) -> int | undefined:
        """
        获取creep的分数
        * 如果creep不存在，则返回undefined
        """
        if not self._creep or not self._creep.exists:
            return undefined
        return self._creep.grade

    @property
    def log(self) -> bool:
        """
        获取是否输出日志
        """
        return self._log

    @property
    def direction(self) -> int | undefined:
        """
        获取creep的诞生方向
        """
        return self._direction

    @direction.setter
    def direction(self, direction: int):
        """
        设置creep的诞生方向
        """
        self._direction = direction

    @property
    def priority(self) -> int:
        """
        获取creep的权重
        """
        return self._priority

    @property
    def recipe(self) -> list[str]:
        """
        获取recipe
        """
        return self._recipe

    @property
    def optimise(self) -> bool:
        """
        获取是否优化recipe
        """
        return self._optimise

    @property
    def extension(self) -> bool:
        """
        获取是否使用extension
        """
        return self._extension

    @property
    def spawnable(self) -> bool:
        """
        获取是否可以孵化
        * 如果一个creep不能孵化，则视作虚拟逻辑。
        * 后续执行的回调函数的第一个参数将是undefined(None)
        * 如果不手动停止，那么将永远不会结束
        """
        return self._spawnable

    @property
    def team(self) -> str:
        """
        获取creep的队伍绑定
        """
        return self._team

    @property
    def parent(self) -> "CreepLogic" | None:
        """
        获取creep的父逻辑(指由那个逻辑link此逻辑的)
        * 如果没有父逻辑，则返回None
        """
        return self._parent

    @property
    def spawn(self) -> st.spawn | undefined:
        """
        获取spawn
        * 如果spawn未绑定，则返回undefined
        """
        return self._spawn

    @property
    def creep(self) -> st.creep | undefined:
        """
        获取creep
        * 如果creep未绑定，则返回undefined
        """
        return self._creep

    @property
    def once(self) -> bool:
        """
        获取once (是否是一次性的creep)
        """
        return self._once

    @property
    def alive(self) -> bool:
        """
        * 检查此对象是否存活 | called by user to check if this object is alive
        """
        return self._alive

    @property
    def enabled(self) -> bool:
        """
        * 检查此对象是否在运行 | called by user to check if this object is running
        """
        return self._enabled

    @property
    def ready(self) -> bool:
        """
        * 检查该logic是否已经完全就绪 | Check if the logic is ready
        * 如果一个logic被标记为spawnable=False，那么它将被视为虚拟逻辑，即使它已经被初始化，也不会被视为就绪

        Returns:
            bool
        """
        return self._alive and (self._team or self._enabled) and st.creep(self._creep)

    def listenFor(self, ename: str, callback: callable, sources: list | None = None):
        """
        * 监听事件 | Listen for events
        * 如果sources为None，则监听所有来源 | If sources is None, listen to all sources

        Args:
            ename: str  # 事件名 | event name
            callback: callable  # 回调函数 | callback function
            sources: list | None  # 来源列表 | source list

        Returns:
            None
        """
        sources = sources if sources else []
        self._eda.listenFor(ename, callback, sources)

    def pushEvent(self, ename: str, data: dict | UsrObject | None = None):
        """
        * 推送事件 | Push events

        Args:
            ename: str
            data: dict | None

        Returns:
            None
        """
        self._eda.pushEvent(ename, data)

    def cancelListen(self, ename: str):
        """
        * 取消监听事件 | Cancel listening for events
        """
        self._eda.cancelListen(ename)

    def schedule(self, delay: int, callback: callable):
        """
        * 延迟执行 | Delay execution

        Args:
            delay: int
            callback: callable
        """
        self._scs.schedule(delay, callback)

    def cancelSchedule(self, callback: callable):
        """
        * 取消延迟执行 | Cancel delay execution

        Args:
            callback: callable
        """
        self._scs.cancelSchedule(callback)

    def link(self, *next_logics: str) -> list["CreepLogic"]:
        """
        * 设置下一个CreepLogic | Set the next CreepLogic
        这个函数会在全局类型表中找到对应NAME的CreepLogic，并将其实例化 |
            This function will find the CreepLogic corresponding to NAME in the global type table and instantiate it

        返回的CreepLogics为disable状态，下一个stage=0的handle会自动启用此。在自动启用前，.enable函数将被禁用 |
            The returned CreepLogics are in disable state, and the next stage=0 handle will automatically enable this. Before automatic enabling, the .enable function will be disabled

        Args:
            next_logics: str  # CreepLogic.NAME

        Returns:
            list[CreepLogic]  # 返回下一个CreepLogic的实例列表 | Returns the instance list of the next CreepLogic
        """
        if len(next_logics) <= 0: return []
        self._next_links.extend(next_logics)

        try:
            return self._enable_next()
        except Exception as e:
            std.error(f"{self.__kind__}.enable_next", e)

    def _enable_next(self) -> list["CreepLogic"]:
        currents = []
        for _type_name in self._next_links:
            _inst = self._scheduler.add(_type_name, self._spawn)
            _inst._spawnable = False
            _inst._enabled = False
            _inst._parent = self
            _inst._auto = True
            self._children.append(_inst)
            currents.append(_inst)
            # _inst.enable(self._spawn)
        self._next_links = []
        return currents

    def relink(self, *next_logics: "CreepLogic"):
        """
        * 重新设置下一个CreepLogic | Reset the next CreepLogic

        Args:
            next_logics: CreepLogic

        Returns:
            None
        """
        for _inst in self._children:
            _inst._parent = self
        self._children.extend(next_logics)

    def refNext(self) -> "CreepLogic" | None:
        """
        * 获取与自身同类型的上一个CreepLogic实例 | Get the previous CreepLogic instance of the same type as itself
        Example:
            比如某个CreepLogic的NAME为"atk"
            atk0.refLast -> None
            atk0.refNext -> atk1
            * 如果atk1寄了并且atk2存在，就返回atk2

        Args:
            None

        Returns:
            None
        """
        lst = self.ref(self._name)
        if lst is None:
            if LANGUAGE == "cn":
                std.warn("CreepLogic.refNext", "指定的基础名不存在: " + self._name)
            else:
                std.warn("CreepLogic.refNext", "The specified base name does not exist: " + self._name)
            return None
        length = len(lst)
        if length <= 1: return None

        found = -1
        for i in range(length - 1):
            if lst[i] == self:
                found = i
                break

        return lst[found + 1] if found != -1 else None

    def refLast(self) -> "CreepLogic" | None:
        """
        * 获取与自身同类型的上一个CreepLogic实例 | Get the previous CreepLogic instance of the same type as itself
        Example:
            比如某个CreepLogic的NAME为"atk"
            atk0.refLast -> None
            atk0.refNext -> atk1
            * 如果atk1寄了并且atk2存在，就返回atk2

        Args:
            None

        Returns:
            None
        """
        lst = self.ref(self._name)
        if lst is None:
            if LANGUAGE == "cn":
                std.warn("CreepLogic.refLast", "指定的基础名不存在: " + self._name)
            else:
                std.warn("CreepLogic.refLast", "The specified base name does not exist: " + self._name)
            return None
        length = len(lst)
        if length <= 1: return None

        found = -1
        for i in range(1, length):
            if lst[i] == self:
                found = i
                break

        return lst[found - 1] if found != -1 else None

    def refAdjacent(self) -> tuple["CreepLogic" | None, "CreepLogic" | None]:
        """
        * 获取与自身同类型的上一个和下一个CreepLogic实例 | Get the previous and next CreepLogic instances of the same type as itself
        Example:
            比如某个CreepLogic的NAME为"atk"
            atk0.refAdjacent -> (None, atk1) | (None, None)

        效率比单独调用refNext和refLast高 | The efficiency is higher than calling refNext and refLast separately

        Args:
            None

        Returns:
            tuple[CreepLogic]
        """
        lst = self.ref(self._name)
        if lst is None:
            if LANGUAGE == "cn":
                std.warn("CreepLogic.refLast", "指定的基础名不存在: " + self._name)
            else:
                std.warn("CreepLogic.refLast", "The specified base name does not exist: " + self._name)
            return None, None
        length = len(lst)
        if length <= 1: return None, None

        found = -1
        for i in range(length):
            if lst[i] == self:
                found = i
                break

        if found == -1:
            return None, None
        elif found == 0:
            return None, lst[1]
        elif found == length - 1:
            return lst[length - 2], None
        else:
            return lst[found - 1], lst[found + 1]

    def enable(self, spawn: StructureSpawn = None):
        if self._auto:
            if LANGUAGE == "cn":
                std.warn(f"{self.name}.enable", "该逻辑已被标记为在下个stage=0的handle自动启用。在此之前enable的功能将被禁用。")
            else:
                std.warn(f"{self.name}.enable",
                         "The logic has been marked to be automatically enabled in the next stage=0 handle. The function enabled before this will be disabled.")
            return

        if not spawn and not self._spawn and not self._creep and not self._once:
            if LANGUAGE == "cn": std.error("CreepLogic.enable", "CreepLogic没有绑定Spawn。请在初始化时或.enable时传入spawn参数。")
            std.error("CreepLogic.enable",
                      "CreepLogic is not bound to Spawn. Please pass the spawn parameter when initializing or .enable.")
        elif spawn:
            self._spawn = spawn

        if not self._enabled:
            self._enabled = True

    def disable(self) -> None:
        """
        * 由用户调用，停用此Updatable | called by user to disable this Updatable

        Args:

        Returns:
            None
        """
        if self._enabled:
            self._enabled = False

    def remove(self) -> None:
        """
        * 由用户调用，停用此Updatable并从Scheduler中移除 | called by user to disable this Updatable and remove it from Scheduler

        Args:

        Returns:
            None
        """
        if self._alive:
            self._alive = False
            self.removed(know)

    def transform(self, new_type_name: str):
        """
        * 由用户调用，将此Updatable转换为新的类型 | called by user to transform this Updatable to a new type

        Example:
            work0.transform("traval-worker")
            # stop work0
            # start traval-worker0

        Args:
            new_type_name: str

        Returns:
            None
        """
        self.remove()
        logic = self._scheduler.add(new_type_name, self._creep)
        if not logic: return
        logic._spawn = self._spawn
        if not self._spawn:
            logic._once = True
            if LANGUAGE == "cn":
                std.warn("CreepLogic.transform", f"CreepLogic没有绑定Spawn。新的Logic:'{new_type_name}'将被强制视作ONCE=True。")
            else:
                std.warn("CreepLogic.transform",
                         f"CreepLogic is not bound to Spawn. The new Logic:'{new_type_name}' will be forced to be ONCE=True.")

    @classmethod
    def productQueue(cls) -> list["CreepLogic"]:
        """
        这个函数可以返回当前等待生产creep的CreepLogic的实例列表 | This function can return a list of instances of CreepLogic waiting to produce creep

        Returns:
            list["CreepLogic"]
        """
        res = []
        sch = Scheduler()
        for logic in sch.updates:
            if st.creep(logic.creep) or (not logic._spawnable and logic._virtual):
                continue
            res.append(logic)
        return res

    @classmethod
    def showQueue(cls) -> None:
        """
        这个函数可以返回当前等待生产creep的CreepLogic的实例列表 | This function can return a list of instances of CreepLogic waiting to produce creep

        Returns:
            list["CreepLogic"]
        """
        queues = cls.productQueue()
        if len(queues) > 0:
            if LANGUAGE == "cn":
                std.info("CreepLogic.showQueue",
                         "当前等待生产creep的CreepLogic的实例列表: \n" + "\n".join([str(lc) for lc in queues]))
            else:
                std.info("CreepLogic.showQueue", "Current list of CreepLogic waiting to produce creep: \n" + "\n".join(
                    [str(lc) for lc in queues]))
        else:
            if LANGUAGE == "cn":
                std.info("CreepLogic.showQueue", "当前没有等待生产creep的CreepLogic的实例列表")
            else:
                std.info("CreepLogic.showQueue", "No CreepLogic waiting to produce creep")
        print()  # 换行

    def jumpQueue(self):
        """
        * 由用户调用，将此CreepLogic插入到队列的最前面 | called by user to insert this CreepLogic into the front of the queue
        """
        queues = self.productQueue()

        if len(queues) > 0:
            # 获取spawn是否忙
            if not st.spawn(self._spawn):
                if LANGUAGE == "cn":
                    std.warn("CreepLogic.jumpQueue", "Not bind a spawn.")
                else:
                    std.warn("CreepLogic.jumpQueue", "Not bind a spawn.")
                return
            if self._spawn.spawning:
                if self._spawn.spawning.creep.name_ == self.name:
                    return  # 已经在生产了 | Already in production
                elif len(queues) > 1:
                    target = queues[
                        1]  # 有多个CreepLogic在等待，插入到第二个 | There are multiple CreepLogic waiting, insert into the second
                else:
                    return  # 只有一个CreepLogic在等待，不插入 | Only one CreepLogic is waiting, do not insert
            else:
                target = queues[0]
        else:
            return

        idx = self._scheduler.updates.indexOf(target)
        if idx >= 0:
            # > insert self.sch.updates.remove(self)
            # > insert self.sch.updates.splice(idx, 0, self)
            pass

    def handle(self, stage: int = 0) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update

        Args:
            stage: int  # 用于标记当前执行的阶段 | Used to mark the current stage of execution
            $ 0: StepBefore
            $ 1: Step
            $ 2: StepAfter
            $ 3: Draw

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """

        # 检查是否死亡 | Check if dead
        if not self._alive:
            return False

        # 检查自动启用 | Check auto enable
        # if self.name == "梦想家0" and stage == 0:
        #     print(self._parent)

        if self._auto and stage == 0 and self._parent and (
                self._parent.ready or (self._parent._virtual and self._parent._enabled)):
            self._spawnable = True
            self._auto = False
            self.enable()

        # 检查队伍状态 | Check team status
        if self._enabled and self._team and self._team._alive:
            self.disable()

        # 检查虚拟状态 | Check virtual status
        if self._virtual:
            self._spawnable = False

        # 检查加载状态 | Check load status
        if not self._loaded and (self._team or self._enabled):
            self._loaded = True
            # loading回调函数
            try:
                self.loading(know, *self._children)
            except Exception as e:
                if LANGUAGE == 'cn':
                    std.error(self.name + ".loading", "执行用户加载逻辑时发生错误: " + str(e))
                else:
                    std.error(self.name + ".loading",
                              "An error occurred while executing the user loading logic: " + str(e))

        # 检查Creep状态
        if self._spawnable:
            if not st.creep(self._creep, False):  # False表明不关心exists
                if not self._recipe:  # 表明这是一个一次性的CreepLogic | This is a one-time CreepLogic
                    self.remove()
                    return False
                elif self._spawn:  # 尝试创建新的Creep | Try to create a new Creep
                    if not self._extension:
                        if get.energy(self._spawn) < self._cost:
                            return False

                    self._creep = self._spawn.create(self._recipe, self.name, self._optimise, self._direction)
                    if st.jnumber(self._creep) and self._creep > 0:
                        self._extension = True  # 重置为True，表示不再限制extension | Reset to True, indicating no longer limiting extension
                    if not st.creep(self._creep):
                        self._creep = None
                        return False
                    self._creep.logic = self
                    self._inited = False
                else:
                    std.error("CreepLogic", "CreepLogic must __init__ with a creep or has recipe and a spawn.")
            elif not self._creep.exists:
                self.killed(know, *self._children)
                self._creep = None
                if self._once: self.remove()
                return False

        # check 是否使能 | check if enabled
        if not self._enabled:
            return False

        # check 是否就绪 | check if ready
        if not self._creep and not self._virtual:
            return False

        # step
        if stage is 0:
            # 执行更新逻辑 | exec Update=
            if not self._inited:
                try:
                    self.birth(self._creep, know, *self._children)
                    self._inited = True
                except Exception as e:
                    std.error(f"{self.__kind__}.init", e)

            # stage
            if self._stagelv == 0 and len(self.__states__):
                self.handle0(self._log)

            try:
                self.step0(self._creep, know, *self._children)
            except Exception as e:
                std.error(f"{self.__kind__}.step0", e)
        elif stage is 1:
            # stage
            if self._stagelv == 1 and len(self.__states__):
                self.handle0(self._log)

            # logic
            try:
                self.step(self._creep, know, *self._children)
            except Exception as e:
                std.error(f"{self.__kind__}.step", e)
        elif stage is 2:
            # stage
            if self._stagelv == 2 and len(self.__states__):
                self.handle0(self._log)

            # logic
            try:
                self.step2(self._creep, know, *self._children)
            except Exception as e:
                std.error(f"{self.__kind__}.step2", e)
        elif stage is 3:
            # stage
            if self._stagelv == 3 and len(self.__states__):
                self.handle0(self._log)

            # logic
            if not self._draw: return True
            self._view.clear()
            try:
                self.draw(self._creep, self._view, know, *self._children)
            except Exception as e:
                std.error(f"{self.__kind__}.draw", e)
        else:
            std.error(f"{self.__kind__}.handle", "Unexpected stage=" + str(stage))

        return True

    BRACKET_COLOR = std.ansi_color(255, 248, 220)
    BRACKET_COLOR2 = std.ansi_color(255, 228, 225)

    @staticmethod
    def _get_state_(logic):
        return "🟢 " if logic.ready else (
            "🔵 " if (logic.enabled or logic.team) else (
                "🟡 " if logic.alive else "☠️"
            )
        )

    @staticmethod
    def _expand_name_(_name, _length):
        # 将name扩充到指定长度，中文视作2字符，英文视作1字符
        # 内联计算字符串的实际长度，中文字符算2，英文字符算1
        current_length = 0
        for char in _name:
            if '\u4e00' <= char <= '\u9fff':  # 判断是否为中文字符
                current_length += 1.6
            else:
                current_length += 1

        current_length = math.ceil(current_length)

        # print("name:", _name, "length:", current_length)

        # 如果当前长度已经大于或等于目标长度，则直接返回
        if current_length >= _length:
            return _name

        # 计算需要填充的空格数量
        spaces_to_add = _length - current_length

        # 在右侧填充空格
        expanded_blanks = []
        for i in range(spaces_to_add):
            expanded_blanks.append(" ")
        expanded_name = _name + "".join(expanded_blanks)

        return expanded_name

    def __str__(self):
        creep_flag = st.creep(self._creep)
        spawn_flag = st.spawn(self._spawn)
        stage_flag = len(self.__states__) > 0
        if creep_flag:
            creep_string = str(self._creep)
        else:
            creep_string = std.wrap_bracket("未绑定Creep" if LANGUAGE == "cn" else "No Creep", "[", self.BRACKET_COLOR2)
        if spawn_flag:
            spawn_string = str(self._spawn)
        else:
            spawn_string = std.wrap_bracket("未绑定Spawn" if LANGUAGE == "cn" else "No Spawn", "[", self.BRACKET_COLOR2)
        if stage_flag:
            lvstr = "⓿" if self._stagelv == 0 else "❶" if self._stagelv == 1 else "❷" if self._stagelv == 2 else "❸"
            stage_string = std.wrap_bracket(lvstr + self.current, "[", self.BRACKET_COLOR)
        else:
            stage_string = ""

        if creep_flag:
            spawn_string = ""
        elif spawn_flag:
            creep_string = ""

        if not self._parent or self._parent._virtual or self._parent.ready:
            after_string = f"{creep_string}{spawn_string}{stage_string}"
        else:
            after_string = std.wrap_bracket(f"{self._get_state_(self._parent)}{self._parent.name}", "[",
                                            self.BRACKET_COLOR)
        return f"{self._get_state_(self)}{self._expand_name} {after_string}"


class VirtualLogic(CreepLogic):
    VIRTUAL = True

    def __str__(self):
        stage_flag = len(self.__states__) > 0
        if stage_flag:
            lvstr = "⓿" if self._stagelv == 0 else "❶" if self._stagelv == 1 else "❷" if self._stagelv == 2 else "❸"
            stage_string = std.wrap_bracket(lvstr + self.current, "[", self.BRACKET_COLOR)
        else:
            stage_string = ""

        return f"{self._get_state_(self)}{self._expand_name} 🌐{stage_string}"


class TeamLogic(VirtualLogic):
    __teams__ = {}

    def _recursive_search(self, logic: CreepLogic) -> list:
        result = []
        result.extend(logic._children)
        for child in logic._children:
            result.extend(self._recursive_search(child))
        return result

    def __init__(self, leader: CreepLogic, *mates: CreepLogic):
        super().__init__()
        self._spawnable = False
        self._leader = leader
        self._mates: list[CreepLogic] = list(mates)

        # 遍历mates的子逻辑, 并添加到mates中
        childs = []
        for mate in self._mates:
            childs.extend(self._recursive_search(mate))
        self._mates.extend(childs)
        # print(self.name, "mates:", self._mates)

        # 检查leader
        if not self._leader:
            if LANGUAGE == "cn":
                std.error("TeamLogic", "TeamLogic必须有一个leader. 传入:", str(leader))
            else:
                std.error("TeamLogic", "TeamLogic must have a leader. Passed in:", str(leader))
        if self._leader.ready:
            self._leader.disable()

        # 检查标记
        if self._leader._team and self._leader._team.alive:
            if LANGUAGE == "cn":
                std.error(f"{self.name}.__init__", f"leader已经加入了一个team: {self._leader._team}")
            else:
                std.error(f"{self.name}.__init__", f"leader has already joined a team: {self._leader._team}")
        for mate in self._mates:
            if mate._team and mate._team.alive:
                if LANGUAGE == "cn":
                    std.error(f"{self.name}.__init__", f"mate:{mate.name}已经加入了一个team: {mate._team}")
                else:
                    std.error(f"{self.name}.__init__", f"mate:{mate.name} has already joined a team: {mate._team}")

        if self.name in TeamLogic.__teams__:
            if LANGUAGE == "cn":
                std.error("TeamLogic", "TeamLogic的name不能重复. 传入:", str(self.name))
            else:
                std.error("TeamLogic", "TeamLogic name must be unique. Passed in:", str(self.name))
        else:
            # 顺带清理一下death的team
            self._clear_dead()
            TeamLogic.__teams__[self.name] = self

        # 标记team
        self._leader._team = self
        for mate in self._mates:
            mate._team = self

        self._spawn = self._leader._spawn
        self._scheduler.add(self)

    @classmethod
    def _clear_dead(cls):
        removes = []
        for k, v in cls.__teams__.items():
            if not v.alive:
                removes.append(k)

        for k in removes:
            del cls.__teams__[k]

    @property
    def leader(self) -> CreepLogic:
        return self._leader

    @property
    def mates(self) -> list[CreepLogic]:
        return self._mates

    def remove(self) -> None:
        super().remove()
        # 清理team
        self._leader._team = None
        self._leader.enable()
        for mate in self._mates:
            mate._team = None
            mate.enable()

    def handle(self, stage: int = 0) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update

        Args:
            stage: int  # 用于标记当前执行的阶段 | Used to mark the current stage of execution
            $ 0: StepBefore
            $ 1: Step
            $ 2: StepAfter
            $ 3: Draw

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """
        # 检查leader
        if not self._leader.alive:
            for mate in self._mates:
                mate.enable()
            self.remove()
            return False

        # 停用leader
        if self._leader.ready:
            self._leader.disable()

        # 检查是否死亡 | Check if dead
        if not self._alive:
            return False

        # 检查自动启用 | Check auto enable
        if self._auto and stage == 0 and self._parent and (
                self._parent.ready or (self._parent._virtual and self._parent._enabled)):
            self._auto = False
            self.enable()

        # 检查leader的状态，决定是恢复mates的独立逻辑还是继续执行team逻辑 | Check the status of the leader to determine whether to resume the independent logic of mates or continue to execute the team logic
        if not self._leader._creep or not self._leader._creep.exists:
            self._leader.enable()
            for mate in self._mates:
                mate.enable()
            return False
        else:
            for mate in self._mates:
                if not mate.ready: continue
                mate.disable()

        # check 是否使能 | check if enabled
        if not self._enabled:
            return False

        # step
        if stage is 0:
            # 执行更新逻辑 | exec Update=
            if not self._inited:
                try:
                    self.birth(self._leader.creep, know, *self._mates)
                    self._inited = True
                except Exception as e:
                    std.error(f"{self.__kind__}.init", e)

            # stage
            if self._stagelv == 0 and len(self.__states__):
                self.handle0(self._log)

            try:
                self.step0(self._creep, know, *self._children)
            except Exception as e:
                std.error(f"{self.__kind__}.step0", e)
        elif stage is 1:
            # stage
            if self._stagelv == 1 and len(self.__states__):
                self.handle0(self._log)

            # logic
            try:
                self.step(self._leader.creep, know, *self._mates)
            except Exception as e:
                std.error(f"{self.__kind__}.step", e)
        elif stage is 2:
            # stage
            if self._stagelv == 2 and len(self.__states__):
                self.handle0(self._log)

            # logic
            try:
                self.step2(self._leader.creep, know, *self._mates)
            except Exception as e:
                std.error(f"{self.__kind__}.step2", e)
        elif stage is 3:
            # stage
            if self._stagelv == 3 and len(self.__states__):
                self.handle0(self._log)

            # logic
            if not self._draw: return True
            self._view.clear()
            try:
                self.draw(self._leader._creep, self._view, know, *self._mates)
            except Exception as e:
                std.error(f"{self.__kind__}.draw", e)
        else:
            std.error(f"{self.__kind__}.handle", "Unexpected stage=" + str(stage))

        return True

    def __str__(self):
        stage_flag = len(self.__states__) > 0
        if stage_flag:
            lvstr = "⓿" if self._stagelv == 0 else "❶" if self._stagelv == 1 else "❷" if self._stagelv == 2 else "❸"
            stage_string = std.wrap_bracket(lvstr + self.current, "[", self.BRACKET_COLOR)
        else:
            stage_string = ""

        return f"{self._get_state_(self)}{self._expand_name} 👥{stage_string} 👑{self._leader.name} 🤺{len(self._mates)}"
