// > sort 2
import {isinstance, LANGUAGE, print, sorted} from "../build/main.mjs";
import {get} from "./get";

const jprint = console.log;

const JS_RED = "\x1b[31m";
const JS_GREEN = "\x1b[32m";
const JS_YELLOW = "\x1b[33m";
const JS_BLUE = "\x1b[34m";
const JS_PURPLE = "\x1b[35m";
const JS_CYAN = "\x1b[36m";
const JS_WHITE = "\x1b[37m";
const JS_GRAY = "\x1b[90m";
const JS_RESET = "\x1b[0m";
const BG_RED = "\x1b[41m";
const BG_GREEN = "\x1b[42m";
const BG_YELLOW = "\x1b[43m";
const BG_BLUE = "\x1b[44m";
const BG_PURPLE = "\x1b[45m";
const BG_CYAN = "\x1b[46m";
const BG_WHITE = "\x1b[47m";
const BG_GRAY = "\x1b[100m";
const BG_RESET = "\x1b[0m";

var ARENA_NAME = arenaInfo["name"];
var ARENA_SHORT_NAME = arenaInfo["name"].toLowerCase().substring(0, 2);
var ARENA_LEVEL = arenaInfo["level"];
var ARENA_SEASON = arenaInfo["season"];
var ARENA_TICKS_LIMIT = arenaInfo["ticksLimit"];
var ARENA_CPU_TIME_LIMIT = arenaInfo["cpuTimeLimit"];
var ARENA_CPU_TIME_LIMIT_FIRST_TICK = arenaInfo["cpuTimeLimitFirstTick"];
const UsrObject = __class__('UsrObject', [object], {
    __module__: __name__,
});

class GlobalKnowledge {
    constructor() {
        this.now = 0
        this.battles = {
            enemies: [],
            friends: [],
        }
        this.civilian = {
            enemies: [],
            friends: [],
        }
        this.clusters = {  // 特指battles的clusters
            enemies: [],
            friends: [],
        }
    }

    handle() {
        this.now = get.ticks();
        const friends = get.friends();
        const enemies = get.enemies();
        this.battles.enemies = get.filter(enemies, c => st.atkable(c) || st.healable(c));
        this.battles.friends = get.filter(friends, c => st.atkable(c) || st.healable(c));
        this.civilian.enemies = get.filter(enemies, c => !st.atkable(c) && !st.healable(c));
        this.civilian.friends = get.filter(friends, c => !st.atkable(c) && !st.healable(c));
        this.clusters.enemies = get.clusters(this.battles.enemies, 4, 1);
        this.clusters.friends = get.clusters(this.battles.friends, 4, 1);

        if (this.view === undefined) this.view = View(1);
        this.view.clear();
        for (const cluster of this.clusters.enemies) {
            if (cluster.grade < 20) continue;
            this.view.cluster(cluster, "#CD5C5C");
        }
        for (const cluster of this.clusters.friends) {
            if (cluster.grade < 20) continue;
            this.view.cluster(cluster, "#87CEEB");
        }
    }
}

const know = new GlobalKnowledge();

const Point = __class__('Point', [object], {
    __module__: __name__,
    exists: true,
    get __init__() {
        return __get__(this, function (self, x, y) {
            self.x = x;
            self.y = y;
        });
    },

    fcenter: function (...points) {
        var length = len(points);
        if (length === 0) {
            return Point(0, 0);
        }
        var x_sum = 0, y_sum = 0;
        for (const p of points) {
            x_sum += p.x;
            y_sum += p.y;
        }
        return Point(x_sum / length, y_sum / length);
    },
    center: function (...points) {
        const p = Point.fcenter(...points);
        p.x = Math.floor(p.x);
        p.y = Math.floor(p.y);
        return p;
    },
    distance: function (target) {
        if (!target || target.x === undefined || target.y === undefined) {
            if (LANGUAGE === "cn") std.warn("Point.distance", `typeof(target)=${typeof target}, 没有.x或者.y属性`, "将返回100");
            else std.warn("Point.distance", `typeof(target)=${typeof target}, no .x or .y property`, "will return 100");
            return 100;
        }
        return Math.max(Math.abs(this.x - target.x), Math.abs(this.y - target.y));

    },
    at: function (target) {
        return this.x === target.x && this.y === target.y;
    },
    nat: function (target) {
        return this.x !== target.x || this.y !== target.y;
    },
    offset: function(dx, dy) {
        return Point(this.x + dx, this.y + dy);
    },
    add: function (other) { return Point(this.x + other.x, this.y + other.y); },
    sub: function (other) { return Point(this.x - other.x, this.y - other.y); },
    neg: function () { return Point(-this.x, -this.y); },
    dot: function (other) { return this.x * other.x + this.y * other.y; },
    mul: function (other) { return Point(this.x * other, this.y * other); },
    crs: function (other) { return this.x * other.y - this.y * other.x; },
    closest: function(objs, filter_fn = undefined, notme = true) {
        if (len(objs) === 0) {
            return null;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.closest`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.closest`, 'No objects left after filtering.')
            return null;
        }
        if (notme) {
            objs = objs.filter(o => o !== this);
        }
        const ret = findClosestByRange(this, objs);
        if (ret === undefined) {
            return null;
        }
        return ret;
    },
    quickest: function(objs, filter_fn = undefined, options = undefined, notme = true) {
        if (len(objs) === 0) {
            return null;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.quickest`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.quickest`, 'No objects left after filtering.')
            return null;
        }
        if (notme) {
            objs = objs.filter(o => o !== this);
        }
        const ret = findClosestByPath(this, objs, options);
        if (ret === undefined) {
            return null;
        }
        return ret;
    },
    inrange: function(objs, range, filter_fn = null) {
        if (len(objs) === 0) {
            return objs;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.inrange`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.inrange`, 'No objects left after filtering.')
            return null;
        }
        const tmp = [];
        for (const _obj of objs) {
            if (this.distance(_obj) <= range) {
                tmp.append(_obj);
            }
        }
        return tmp;
    },

    toString: function () {
        return `(${this.x}◈${this.y})`;
    }
});

var P = Point;

const Rect = __class__('Rect', [object], {
    __module__: __name__,
    exists: true,
    get __init__() {
        return __get__(this, function (self, lt, rb_w, h = undefined, ltrb = undefined) {

            if (ltrb) {
                if (!isinstance(lt, int) ||
                    !isinstance(rb_w, int) ||
                    !isinstance(h, int) ||
                    !isinstance(ltrb, int)) {
                    if (LANGUAGE === "cn") std.error("Rect.init", "lt, rb_w, h, ltrb must be number. Got:", str(lt), str(rb_w), str(h), str(ltrb));
                    else std.error("Rect.init", "lt, rb_w, h, ltrb must be number. Got:", str(lt), str(rb_w), str(h), str(ltrb));
                }
                lt = {x: lt, y: rb_w};
                rb_w = {x: h, y: ltrb};
                h = undefined;
            } else if (!lt || lt.x === undefined || lt.y === undefined) {
                if (LANGUAGE === "cn") std.error("Rect.init", "lt must be Point. Got:", str(lt));
                else std.error("Rect.init", "lt must be Point. Got:", str(lt));
            }
            self.left = Math.floor(lt.x);
            self.top = Math.floor(lt.y);
            if ((!rb_w || rb_w.x === undefined || rb_w.y === undefined) && !isinstance(rb_w, int)) {
                if (LANGUAGE === "cn") std.error("Rect.init", "rb_w must be Point or number. Got:", str(rb_w));
                else std.error("Rect.init", "rb_w must be Point or number. Got:", str(rb_w));
            }
            if (isinstance(rb_w, int)) {
                if (!isinstance(h, int)) {
                    if (LANGUAGE === "cn") std.error("Rect.init", "h must be number. Got:", str(h));
                    else std.error("Rect.init", "h must be number. Got:", str(h));
                }
                self.right = Math.floor(lt.x + rb_w);
                self.bottom = Math.floor(lt.y + h);
            } else {
                self.right = Math.floor(rb_w.x);
                self.bottom = Math.floor(rb_w.y);
            }
        });
    },
    offset(dx, dy) {
        dx = Math.floor(dx);
        dy = Math.floor(dy);
        return Rect(Point(this.left + dx, this.top + dy), Point(this.right + dx, this.bottom + dy));
    },
    get center() {
        return Point((this.left + this.right) / 2, (this.top + this.bottom) / 2);
    },
    get width() {
        return this.right - this.left;
    },
    get height() {
        return this.bottom - this.top;
    },
    get area() {
        return this.width * this.height;
    },
    get __str__() {
        return __get__(this, function (self) {
            return 'Rect(left=' + self.left + ', top=' + self.top + ', right=' + self.right + ', bottom=' + self.bottom + ')';
        });
    },
});

class ListenTask {
    constructor(name, func, ...sources) {
        /**
         * @param func {function}
         * @param sources {string[]}
         * */
        if (!func) {
            if (LANGUAGE === "cn") std.error('ListenTask.constructor', 'func 参数不能为空');
            else std.error('ListenTask.constructor', 'func cannot be empty');
        }
        this.name = name;
        this.func = func;
        this.sources = sources;
        for (let i = 0; i < this.sources.length; i++) {
            if (typeof this.sources[i] !== 'string') {
                this.sources[i] = this.sources[i].name;
            }
        }
    }

    checkMatch(target_name) {
        /**
         * @description 检查目标名称是否匹配this.sources。检查使用的是字符串的startswith方法。
         * @example aa00 is matched by [aa, bb]
         * @param target_name {string}
         * */
        if (!target_name) {
            if (LANGUAGE === "cn") std.error('ListenTask.checkMatch', 'target_name 参数不能为空');
            else std.error('ListenTask.checkMatch', 'target_name cannot be empty');
        }
        if (this.sources.length === 0) {
            return true;
        }
        for (const src of this.sources) {
            if (target_name.startsWith(src)) {
                return true;
            }
        }
        return false;
    }

    tryCall(name, ename, ...args) {
        try {
            this.func(...args);
        } catch (e) {
            if (LANGUAGE === "cn") std.warn("ListenSystem", `在 ${name} 的监听事件 ${ename} 中出错:`, e);
            else std.warn("ListenSystem", `Error in ${name}.listen-for ${ename}:`, e);
            // throw e;
        }

    }
}

class ListenSystem {
    constructor(game_object, sp_name=undefined) {
        if (!game_object) {
            if (LANGUAGE === "cn") std.error('ListenSystem.constructor', 'game_object 不能为空');
            else std.error('ListenSystem.constructor', 'game_object cannot be empty');
        }
        this.object = game_object;
        if (sp_name){
            this.name = sp_name;
        } else {
            this.name = this.object.name ? this.object.name : this.object.id;
        }
        print(this.name);
        if (typeof this.name !== 'string') {
            if (LANGUAGE === "cn") std.error('ListenSystem.constructor', 'name 必须是字符串');
            else std.error('ListenSystem.constructor', 'name must be a string');
        }
    }

    static __listens__ = new Map(); // 使用Map代替对象存储事件监听

    listenFor(ename, func, sources = undefined) {
        if (typeof ename !== 'string') {
            if (LANGUAGE === "cn") std.error('ListenSystem.listenFor', 'ename 必须是字符串');
            else std.error('ListenSystem.listenFor', 'ename must be a string');
        }
        if (typeof func !== 'function') {
            if (LANGUAGE === "cn") std.error('ListenSystem.listenFor', 'func 必须是函数');
            else std.error('ListenSystem.listenFor', 'func must be a function');
        }
        if (!this.name) {
            if (LANGUAGE === "cn") std.error('ListenSystem.listenFor', 'name 不能为空');
            else std.error('ListenSystem.listenFor', 'name cannot be empty');
        }
        if (sources === undefined) sources = [];
        if (!ListenSystem.__listens__.has(ename)) {
            ListenSystem.__listens__.set(ename, new Map());
        }
        ListenSystem.__listens__.get(ename).set(this.name,
            new ListenTask(this.name, func, ...sources)
        );
    }

    cancelListen(ename) {
        if (typeof ename !== 'string') {
            if (LANGUAGE === "cn") std.error('ListenSystem.cancelListen', 'ename 必须是字符串');
            else std.error('ListenSystem.cancelListen', 'ename must be a string');
        }
        if (!this.name) {
            if (LANGUAGE === "cn") std.error('ListenSystem.cancelListen', 'name 不能为空');
            else std.error('ListenSystem.cancelListen', 'name cannot be empty');
        }
        if (ListenSystem.__listens__.has(ename)) {
            const clstype = ListenSystem.__listens__.get(ename);
            if (clstype.has(this.name)) {
                clstype.delete(this.name);
                if (clstype.size === 0) {
                    ListenSystem.__listens__.delete(ename);
                }
            } else {
                if (LANGUAGE === "cn") console.warn(`没有名为 ${this.name} 的监听器监听事件 ${ename}`);
                else console.warn(`No listener with name ${this.name} for event ${ename}`);
            }
        } else {
            if (LANGUAGE === "cn") console.warn(`没有名为 ${ename} 的事件`);
            else console.warn(`No event named ${ename}`);
        }
    }

    pushEvent(ename, data=null) {
        if (typeof ename !== 'string') {
            if (LANGUAGE === "cn") std.error('ListenSystem.pushEvent', 'ename 必须是字符串');
            else std.error('ListenSystem.pushEvent', 'ename must be a string');
        }
        if (typeof data !== 'object' && data !== null) {
            if (LANGUAGE === "cn") std.error('ListenSystem.pushEvent', 'data 必须是对象');
            else std.error('ListenSystem.pushEvent', 'data must be an object');
        }
        if (!this.name) {
            if (LANGUAGE === "cn") std.error('ListenSystem.pushEvent', 'name 不能为空');
            else std.error('ListenSystem.pushEvent', 'name cannot be empty');
        }
        if (!data) data = {};
        data.py_name = this.name;
        data.inst = this.object;
        if (ListenSystem.__listens__.has(ename)) {
            const listeners = ListenSystem.__listens__.get(ename);
            for (const [name, task] of listeners) {
                if (task.checkMatch(this.name)) {
                    task.tryCall(name, ename, data);
                }
            }
        }
    }
}

class ScheduleSystem {
    static __schedules__ = new Map();  // str(tick) : [callback, ...]

    constructor(gobject) {
        this.object = gobject;
        this.name = this.object.name ? this.object.name : this.object.id;
        if (typeof this.name !== 'string') {
            if (LANGUAGE === "cn") std.error('ScheduleSystem.constructor', 'name 必须是字符串');
            else std.error('ScheduleSystem.constructor', 'name must be a string');
        }
    }

    schedule(tick, callback) { // tick: int
        if (typeof tick !== 'number') {
            if (LANGUAGE === "cn") std.error('ScheduleSystem.schedule', 'tick 必须是数字');
            else std.error('ScheduleSystem.schedule', 'tick must be a number');
        }
        const str_tick = tick.toString();
        let callbacks = ScheduleSystem.__schedules__.get(str_tick);
        if (!callbacks) {
            callbacks = [];
            ScheduleSystem.__schedules__.set(str_tick, callbacks);
        }
        callbacks.push(callback);
    }

    cancelSchedule(callback) {
        for (const [tick, callbacks] of ScheduleSystem.__schedules__) {
            const index = callbacks.indexOf(callback);
            if (index !== -1) {
                callbacks.splice(index, 1);
            }
            if (callbacks.length === 0) {
                ScheduleSystem.__schedules__.delete(tick);
            }
        }
    }

    handle() {
        const s_tick = know.now.toString();
        const callbacks = ScheduleSystem.__schedules__.get(s_tick);
        if (callbacks) {
            for (const callback of callbacks) {
                try {
                    callback();
                } catch (e) {
                    if (LANGUAGE === "cn") std.warn("ScheduleSystem", `在 ${this.name} 的调度任务中出错:`, e);
                    else std.warn("ScheduleSystem", `Error in ${this.name}'s schedule task:`, e);
                }
            }
            ScheduleSystem.__schedules__.delete(s_tick);
        }
    }
}


class PartsVector {
    static INDEXS = {
        [MOVE]: 0,
        [CARRY]: 1,
        [WORK]: 2,
        [ATTACK]: 3,
        [RANGED_ATTACK]: 4,
        [HEAL]: 5,
        [TOUGH]: 6
    };

    constructor(named_recipe) {
        this.recipe = (named_recipe instanceof NamedRecipe) ? named_recipe.recipe : named_recipe;
        this.vec7 = [0, 0, 0, 0, 0, 0, 0];

        for (const part of this.recipe) {
            this.vec7[PartsVector.INDEXS[part]] += 1;
        }

        // non-move  non-carry  count
        this.nmCount = this.vec7[2] + this.vec7[3] + this.vec7[4] + this.vec7[5] + this.vec7[6];

        // total count
        this.bodyCount = this.recipe.length;
    }

    add(parts_vector_recipe) {
        if (!(parts_vector_recipe instanceof PartsVector)) {
            parts_vector_recipe = new PartsVector(parts_vector_recipe);
        }
        for (let i = 0; i < this.vec7.length; i++) {
            this.vec7[i] += parts_vector_recipe.vec7[i];
        }
    }

    sub(parts_vector_recipe) {
        if (!(parts_vector_recipe instanceof PartsVector)) {
            parts_vector_recipe = new PartsVector(parts_vector_recipe);
        }
        for (let i = 0; i < this.vec7.length; i++) {
            this.vec7[i] = Math.max(0, this.vec7[i] - parts_vector_recipe.vec7[i]);
        }
    }

    toString() {
        return this.vec7.toString();
    }

    // getter
    get others() {
        return this.nmCount;
    }

    get moves() {
        return this.vec7[0];
    }

    get carries() {
        return this.vec7[1];
    }

    get works() {
        return this.vec7[2];
    }

    get melees() {
        return this.vec7[3];
    }

    get ranges() {
        return this.vec7[4];
    }

    get heals() {
        return this.vec7[5];
    }

    get toughs() {
        return this.vec7[6];
    }

    static similarity(a, b) {
        if (!a || !b) {
            throw new Error("PartsVector instances cannot be None");
        }

        const a_vec7 = a.vec7;
        const b_vec7 = b.vec7;

        if (a_vec7.length !== b_vec7.length) {
            throw new Error("Vectors must be of the same length, but:" + a_vec7.length.toString() + " and " + b_vec7.length.toString());
        }

        let dotProduct = 0;
        let normA = 0;
        let normB = 0;

        for (let i = 0; i < a_vec7.length; i++) {
            dotProduct += a_vec7[i] * b_vec7[i];
            normA += a_vec7[i] ** 2;
            normB += b_vec7[i] ** 2;
        }

        normA = Math.sqrt(normA);
        normB = Math.sqrt(normB);

        if (normA === 0 || normB === 0) {
            return 0.0;
        }

        return dotProduct / (normA * normB);
    }

    static selectVector(a, vecs) {
        if (!a || !Array.isArray(vecs)) {
            throw new Error("Invalid input for selectVector:" + typeof a + " and " + typeof vecs);
        }

        return vecs.reduce((maxVec, currentVec) => {
            return PartsVector.similarity(a, currentVec) > PartsVector.similarity(a, maxVec) ? currentVec : maxVec;
        });
    }

    static parts(creep, broken = true) {
        if (!(creep) || !(creep.body)) {
            if (LANGUAGE === "cn") std.warn('get.parts', 'creep为空或没有body。返回空列表: typeof creep:', typeof creep);
            else std.warn('get.parts', 'creep is NA or has no body. Return empty list: typeof creep', typeof creep);
            return [];
        }
        // print(creep.body);
        const prts = [];
        for (const prt of creep.body) {
            if (broken || prt.hits > 0) {
                prts.append(prt.type);
            }
        }
        return prts;
    }

    static SCORE_TABLE = {
        [ATTACK]: 10,
        [RANGED_ATTACK]: 15,
        [HEAL]: 15,
        [TOUGH]: 1,
        [MOVE]: 2,
        [CARRY]: 2,
        [WORK]: 5,
    }
    static partsGrade(creep_recipe) {
        let carries = 0;
        let moves = 0;
        let score = 0;
        const prts = (creep_recipe instanceof Creep) ? creep_recipe.recipe : creep_recipe;
        for (let [i, prt] of enumerate(prts)) {
            if (prt === ATTACK) {
                score += 10 + i / 5;
            } else if (prt === RANGED_ATTACK) {
                score += 15 + i / 4;
            } else if (prt === HEAL) {
                score += 15 + i;
            } else {
                score++;
                if (prt === MOVE) {
                    moves++;
                } else if (prt === CARRY) {
                    carries++;
                }
            }
        }
        const length = (len(prts) - carries) - moves;
        const movements = moves * 2;
        const swamp_ratio = math.clamp(get.SWAMP_RATIO, 0.1, 0.5);
        // const move_cost = length * (swamp_ratio * 10 + 5);
        const move_cost = length * swamp_ratio * 20;
        let coef;
        if (length === 0) {
            coef = 0;
        } else if (movements > move_cost) {
            coef = 2;
        } else {
            coef = (2 * movements) / move_cost;
        }
        const finalCoef = (1 + coef) / 2;
        return Math.floor(score * finalCoef);
    }
    static partsOptimise(parts) {
        if (len(parts) <= 1) {
            return parts;
        }
        // step1: 将所有的TOUGH部件放在最前面
        const tough_head = parts.filter(part => part === TOUGH);
        const rest_parts = parts.filter(part => part !== TOUGH);

        // step2 create order
        let order = [...new Set(rest_parts.reverse())];

        // step3 count 最少个数的种类
        const each_count = {};
        for (const kind of order) each_count[kind] = rest_parts.filter(part => part === kind).length;
        let min_count = Infinity;
        let min_kind = null;
        for (const kind of order)
            if (each_count[kind] < min_count) {
                min_count = each_count[kind];
                min_kind = kind;
            }

        // step4: 创建 unit_count = { each_count[k].count // min_count } 和 other_count = { each_count[k].count % min_count }
        const unit_count = {};
        const other_count = {};
        for (const kind of order) {
            unit_count[kind] = Math.floor(each_count[kind] / min_count);
            other_count[kind] = each_count[kind] % min_count;
        }

        // step5: create group_pattern. iter 'order', fetch same kind from unit_count if has. until all unit_count is 0
        let group_pattern = [];
        while (true) {
            let flag = false;  // any push will set as true
            for (const kind of order) {
                if (unit_count[kind] > 0) {
                    group_pattern.push(kind);
                    unit_count[kind] -= 1;
                    flag = true;
                }
            }
            if (!flag) break;
        }
        group_pattern = group_pattern.reverse();

        // step6: create others_sequence. iter 'order', fetch same kind from other_count if has. until all other_count is 0
        const others_sequence = [];
        while (true) {
            let flag = false;  // any push will set as true
            for (const kind of order) {
                if (other_count[kind] > 0) {
                    others_sequence.push(kind);
                    other_count[kind] -= 1;
                    flag = true;
                }
            }
            if (!flag) break;
        }

        // inner function: split others into min_count list. can be empty if not has enough
        const split_others = function (min_count, others_sequence) {
            if (min_count <= 1) return [others_sequence];
            const total = (min_count + 1) * min_count / 2;
            let last_idx = 0;
            let cur_idx = 0;
            const res = [];
            for (let i = min_count; i > 0; i--) {
                cur_idx = last_idx + (i / total) * others_sequence.length;
                last_idx = Math.ceil(last_idx);
                cur_idx = Math.ceil(cur_idx);
                res.push(others_sequence.slice(last_idx, cur_idx));
                last_idx = cur_idx;
            }
            return res;
        }

        // step7: split others_sequence into min_count list
        const others_splits = split_others(min_count, others_sequence);

        // step: special optimise:
        // ignore carry, if MOVE * 5  < others: ...
        const tails = [];
        const move_count = each_count[MOVE];
        let not_move_count = 0;
        for (const kind of order) {
            if (kind === MOVE) continue;
            if (kind === CARRY) continue;
            not_move_count += each_count[kind];
        }
        if ((move_count * 2) < (not_move_count * 5)) {
            // 将others_splits[0]中的所有MOVE部件移动到tails中
            const moves = others_splits[0].filter(part => part === MOVE);
            tails.push(...moves);
            others_splits[0] = others_splits[0].filter(part => part !== MOVE);
        } else if (move_count < not_move_count * 5) {
            // 将others_splits[0]中的半数(向上取整)MOVE部件移动到tails中
            const moves = others_splits[0].filter(part => part === MOVE);
            const move_count = Math.ceil(moves.length / 2);
            tails.push(...moves.slice(0, move_count));
            for (let i = 0; i < move_count; i++)
                others_splits[0].remove(MOVE);
        }

        // step8: merge HEAD + (others_splits[i] + group_pattern) + (others_splits[i+1] + group_pattern) + ...
        let res = [];
        res.push(...tough_head);
        for (let i = 0; i < min_count; i++) {
            res.push(...others_splits[i]);
            res.push(...group_pattern);
        }
        res.push(...tails);


        // 提出所有HEAL部件，但是保留最后一个(如果有的话)
        const heals = res.filter(part => part === HEAL);
        if (heals.length > 1) {
            heals.shift();
            for (const heal of heals) res.remove(heal);
            // 找到最后一个HEAL部件，将其余的HEAL部件插入在此
            const last_heal = res.indexOf(HEAL);
            res.splice(last_heal, 0, ...heals);
        }

        return res;
    }
    static PARTS_COST = {
        [MOVE]: 50,
        [WORK]: 100,
        [CARRY]: 50,
        [ATTACK]: 80,
        [RANGED_ATTACK]: 150,
        [HEAL]: 250,
        [TOUGH]: 10,
    }
    static partsCost(creep_recipe) {
        if (creep_recipe instanceof Creep) {
            creep_recipe = creep_recipe.recipe;
        }
        let cost = 0;
        for (const part of creep_recipe) {
            const this_cost = PartsVector.PARTS_COST[part];
            cost += this_cost!==undefined ? this_cost : 0;
        }
        return cost;
    }
}


class NamedRecipe {
    static __recipes__ = new Map();
    static PART_COLOR = {
        [MOVE]: "#a5b7c6",
        [CARRY]: "#5c6e74",
        [WORK]: "#ffdb5b",
        [ATTACK]: "#f92c2e",
        [RANGED_ATTACK]: "#4c8090",
        [HEAL]: "#65d833",
        [TOUGH]: "#c1f0f8"
    }
    static PART_SHORTNAME = {
        [MOVE]: "M",
        [CARRY]: "C",
        [WORK]: "W",
        [ATTACK]: "A",
        [RANGED_ATTACK]: "R",
        [HEAL]: "H",
        [TOUGH]: "T"
    }

    constructor(name, creep_recipe) {
        // check repeat
        if (NamedRecipe.__recipes__.has(name)) {
            if (LANGUAGE === "cn") std.error('NamedRecipe', '重复的配方名: ', name);
            else std.error('NamedRecipe', 'Duplicate recipe name: ', name);
        }

        this.name = name;
        this.recipe = (creep_recipe instanceof GameCreep) ? PartsVector.parts(creep_recipe) : creep_recipe;
        this.grade = PartsVector.partsGrade(this.recipe);
        this.cost = PartsVector.partsCost(this.recipe);
        this.effect = round(this.grade / this.cost * 100, 2) / 100;

        // 创建彩色的配方字符串 使用std.ansi_color() -> str
        this.recipeString = this.recipe.map(part => {
            if (NamedRecipe.PART_SHORTNAME.hasOwnProperty(part) && NamedRecipe.PART_COLOR.hasOwnProperty(part)) {
                const partName = NamedRecipe.PART_SHORTNAME[part];
                const color = NamedRecipe.PART_COLOR[part];
                return std.ansi_color(color) + partName + JS_RESET;
            } else {
                return std.ansi_color('#FF00A0') + part + JS_RESET;
            }
        }).join('');

        // 创建recipeShort字符串，比如M4W2A2H2C2T2
        const vec7 = [0, 0, 0, 0, 0, 0, 0];
        for (const part of this.recipe) {
            vec7[PartsVector.INDEXS[part]] += 1;
        }
        this.recipeShort = vec7.map((count, index) => {
            if (count > 0) {
                return NamedRecipe.PART_SHORTNAME[index] + count.toString();
            }
        }).filter(Boolean).join('');
    }
}

class CreepInfo {
    static ATTACK_POWER = 30;
    static HEAL_POWER = 12;
    static RANGED_ATTACK_POWER = 10;

    constructor(creep) {
        this.creep = creep;
        this.name = creep.name;
        this.py_name = creep.py_name;
        this.namedRecipe = new NamedRecipe(this.name, creep);
        this.partsVector = new PartsVector(this.namedRecipe);
        this.grade = this.namedRecipe.grade;
        this.cost = this.namedRecipe.cost;
        this.effect = this.namedRecipe.effect;

        // 特性描述
        this.melee = st.melee(creep);
        this.ranged = st.ranged(creep);
        this.heal = st.healable(creep);
        this.work = st.workable(creep);
        this.storable = st.storable(creep);

        // 动态特性(确保某个tick内只计算一次)
        this._dynamicUpdateTick = 0;
        this._dynamicPartsVector = null;
    }

    // dynamic
    get dynamicPartsVector() {
        if (know.now !== this._dynamicUpdateTick) {
            this._dynamicUpdateTick = know.now;
            this._dynamicPartsVector = new PartsVector(PartsVector.parts(this.creep, false));
        }
        return this._dynamicPartsVector;
    }

    // 攻击力
    get attackPower() {
        return this.partsVector.melees * CreepInfo.ATTACK_POWER + this.partsVector.ranges * CreepInfo.RANGED_ATTACK_POWER;
    }

    get dynamicAttackPower() {
        return this.dynamicPartsVector.melees * CreepInfo.ATTACK_POWER + this.dynamicPartsVector.ranges * CreepInfo.RANGED_ATTACK_POWER;
    }

    // 近战力
    get meleePower() {
        return this.partsVector.melees * CreepInfo.ATTACK_POWER;
    }

    get dynamicMeleePower() {
        return this.dynamicPartsVector.melees * CreepInfo.ATTACK_POWER;
    }

    // 远程攻击力
    get rangedPower() {
        return this.partsVector.ranges * CreepInfo.RANGED_ATTACK_POWER;
    }

    get dynamicRangedPower() {
        return this.dynamicPartsVector.ranges * CreepInfo.RANGED_ATTACK_POWER;
    }

    // 治疗力
    get healPower() {
        return this.partsVector.heals * CreepInfo.HEAL_POWER;
    }

    get dynamicHealPower() {
        return this.dynamicPartsVector.heals * CreepInfo.HEAL_POWER;
    }

    // 移动率
    get motionAbility() {
        if (this.partsVector.nmCount === 0) return 1;
        return (this.partsVector.moves * 2) / this.partsVector.nmCount / 10;
    }

    get dynamicMotionAbility() {
        if (this.dynamicPartsVector.nmCount === 0) return 1;
        return (this.dynamicPartsVector.moves * 2) / this.dynamicPartsVector.nmCount / 10;
    }

    // 装甲率
    get armorRatio() {
        return round(this.partsVector.toughs / this.partsVector.bodyCount * 100, 2) / 100;
    }

    get dynamicArmorRatio() {
        return round(this.dynamicPartsVector.toughs / this.dynamicPartsVector.bodyCount * 100, 2) / 100;
    }

    __bool_string__(bool) {
        if (bool) return `${JS_GREEN}✔${JS_RESET}`;
        else return `${JS_RED}✘${JS_RESET}`;
    }

    // toString 方法
    toString() {
        if (LANGUAGE === "cn") return `Creep信息(名称: ${this.name}, 配方: ${this.namedRecipe.recipeShort}, 评分: ${this.grade}, 攻击力: ${this.attackPower}, 治疗力: ${this.healPower})`;
        return `CreepInfo(name: ${this.name}, recipe: ${this.namedRecipe.recipeShort}, grade: ${this.grade}, attackPower: ${this.attackPower}, healPower: ${this.healPower})`;
    }

    // details 方法
    details() {
        if (LANGUAGE === "cn") {
            return `Creep详细信息：
	- 名称: ${this.name}
	- 配方: ${this.namedRecipe.recipeString}
	- 评分: ${this.grade} (沼泽率: ${round(get.SWAMP_RATIO * 100, 2)}%)
	- 成本: ${this.cost} Pt
	- 效率: ${round(this.effect * 100, 2)}%
	- 部件向量: ${this.partsVector}
	- 动态部件向量: ${this.dynamicPartsVector}
	- 攻击力: ${this.attackPower}(动态: ${this.dynamicAttackPower})
	- 近战力: ${this.meleePower}(动态: ${this.dynamicMeleePower})
	- 远程攻击力: ${this.rangedPower}(动态: ${this.dynamicRangedPower})
	- 治疗力: ${this.healPower}(动态: ${this.dynamicHealPower})
	- 移动率: ${this.motionAbility * 100}%(动态: ${this.dynamicMotionAbility * 100}%)
	- 装甲率: ${this.armorRatio * 100}%(动态: ${this.dynamicArmorRatio * 100}%)
	- 特性: 近战: ${this.__bool_string__(this.melee)}, 远程: ${this.__bool_string__(this.ranged)}, 治疗: ${this.__bool_string__(this.heal)}, 工作: ${this.__bool_string__(this.work)}, 存储: ${this.__bool_string__(this.storable)}`;
        } else {
            return `CreepInfo Details:
	- Name: ${this.name}
	- Recipe: ${this.namedRecipe.recipeString}
	- Grade: ${this.grade} (Swamp Ratio: ${round(get.SWAMP_RATIO * 100, 2)}%)
	- Cost: ${this.cost} Pt
	- Effect: ${round(this.effect * 100, 2)}%
	- PartsVector: ${this.partsVector}
	- DynamicPartsVector: ${this.dynamicPartsVector}
	- AttackPower: ${this.attackPower} (Dynamic: ${this.dynamicAttackPower})
	- MeleePower: ${this.meleePower} (Dynamic: ${this.dynamicMeleePower})
	- RangedPower: ${this.rangedPower} (Dynamic: ${this.dynamicRangedPower})
	- HealPower: ${this.healPower} (Dynamic: ${this.dynamicHealPower})
	- MotionAbility: ${this.motionAbility * 100}% (Dynamic: ${this.dynamicMotionAbility * 100}%)
	- ArmorRatio: ${this.armorRatio * 100}% (Dynamic: ${this.dynamicArmorRatio * 100}%)
	- Characteristics: melee: ${this.__bool_string__(this.melee)}, ranged: ${this.__bool_string__(this.ranged)}, heal: ${this.__bool_string__(this.heal)}, work: ${this.__bool_string__(this.work)}, storable: ${this.__bool_string__(this.storable)}`;
        }
    }
}

class CreepMotion {
    /**
     * 对己方队友，记录最近一次移动的操作
     * 对敌方敌人，追踪并预测其位置
     * */
    static SEPARATE_LENGTH = 8;  // 每此长度的便取一个点，作为阶段性移动的目标点
    constructor(creep) {
        this.creep = creep

        // 己方运动信息
        this.path = [];  // 记录已有的一组路径，让creep沿着这条路径移动
        this.path_goal = {x: creep.x, y: creep.y};  // 记录路径的终点
        this.path_index = 0;  // 记录路径的索引
        this.next_force = 0;  // 记录下一次强制更新path的时间
        this.last_options = undefined;  // 记录上一次寻路的选项
        /*
        * 假设某个creep需要移动到目标点:
        * 1. this.path为空，那我们为其计算一条路径
        * 	1.1 使用const path_res = searchPath (unit:{x,y}, to:{x, y}, options); 计算一条路径，是一组点
        * 	1.2 根据SEPARETE_LENGTH，取出一些点，作为每个阶段的目标点
        *   1.3 如果path_res长度小于SEPARATE_LENGTH，那么直接单步移动而不保存到path
        *   1.4 如果path_res长度大于SEPARATE_LENGTH，那么保存关键点到path
        *   1.5 同时更新path_goal为path_res的最后一个点;跟新path_index为0
        * 2. this.path不为空，那么取出path[path_index]
        * 	2.1 判断creep与path[path_index]的距离，如果<=1，那么path_index++，如果path_index >= path.length，那么目标点就设为path_goal
        *   2.2 得到阶段性的目标点后，计算creep到阶段点的路径, 作为creep的下一个移动目标
         */

        // 敌方运动信息
        this.lastPoint = {x: this.creep.x, y: this.creep.y};  // 记录敌方creep上一帧的位置

        // debug
        // this.view = View(10);
    }

    seekNext(options = undefined) {
        /**
         * 对友军，寻找下一个目标点
         * 对敌方，寻找下一个预测点
         *
         * @param options {object} 只对友军有效, 作为没有路径时的寻路选项
         * */
        if (options !== undefined) this.last_options = options;
        if (!this.creep.my) {
            if (this.creep.fatigue > 0) return {x: this.creep.x, y: this.creep.y};
            return {x: this.creep.x * 2 - this.lastPoint.x, y: this.creep.y * 2 - this.lastPoint.y};
        } else if (this.creep.x === this.path_goal.x && this.creep.y === this.path_goal.y) {
            return {x: this.creep.x, y: this.creep.y};
        } else {
            let stage_point = this.path_goal;
            if (this.path.length) {
                if (this.path_index < this.path.length) {  // 假设path检查是提前进行的，如果对path不满意，就清空它
                    stage_point = this.path[this.path_index];
                }

                // 如果creep与stage_point距离小于等于1，那么就更新path_index
                if (this.creep.__py__.distance(stage_point) <= 2) {
                    this.path_index++;
                    stage_point = this.path_index >= this.path.length ? this.path_goal : this.path[this.path_index];
                }
            }


            if (!options) options = {};
            if(!options.maxOps) options.maxOps = CreepMotion.SEPARATE_LENGTH * CreepMotion.SEPARATE_LENGTH * 4;
            const path_res = searchPath(this.creep, stage_point, options);
            if (path_res.incomplete) {
                if (LANGUAGE === "cn") std.warn('CreepMotion.seekNext', `${this.creep.name}到目标(${this.path_goal.x}, ${this.path_goal.y})的路径不完整。`);
                else std.warn('CreepMotion.seekNext', `The path from ${this.creep.name} to (${this.path_goal.x}, ${this.path_goal.y}) is incomplete.`);
            }
            return path_res.path[0];
        }
    }

    seekNextStagePoint(new_path_goal, options = undefined) {
        /*
        * 只对友军有效, 寻找下一个阶段性目标点
        * */

        if (options !== undefined) this.last_options = options;
        if (!this.creep.my) {
            if (LANGUAGE === "cn") std.warn('CreepMotion.seekNextStagePoint', '只对己方creep有效:id=', this.creep.id);
            else std.warn('CreepMotion.seekNextStagePoint', 'Only for friendly creeps, id=', this.creep.id);
        }
        if (this.creep.x === new_path_goal.x && this.creep.y === new_path_goal.y) {
            return {x: this.creep.x, y: this.creep.y};
        } else {
            const dist = Math.max(Math.abs(this.creep.x - new_path_goal.x), Math.abs(this.creep.y - new_path_goal.y));
            if (
                (this.next_force < know.now) ||
                (dist <= 20 &&(new_path_goal.x !== this.path_goal.x || new_path_goal.y !== this.path_goal.y)) ||
                (dist > 20 && (Math.abs(new_path_goal.x - this.path_goal.x) > 1 || Math.abs(new_path_goal.y - this.path_goal.y) > 1))
            ) {
                this.path = [];
                this.next_force = know.now + 10;
                this.path_index = 0;
                this.path_goal = {x: new_path_goal.x, y: new_path_goal.y};
                const cell_cost = DEFAULT_MOTION.costMatrix.get(new_path_goal.x, new_path_goal.y);

                const path_res = searchPath(this.creep, cell_cost < 255 ? new_path_goal : {pos: new_path_goal, range: 1}, options);
                if (path_res.incomplete) {
                    if (LANGUAGE === "cn") std.warn('CreepMotion.seekNextStagePoint', `${this.creep.name}到目标(${this.path_goal.x}, ${this.path_goal.y})的路径不完整。`);
                    else std.warn('CreepMotion.seekNextStagePoint', `The path from ${this.creep.name} to (${this.path_goal.x}, ${this.path_goal.y}) is incomplete.`);
                }

                // if (options && options.swampCost){
                //     console.log("options:", options);
                //     console.log("path:", path_res.path);
                // }

                // 每隔SEPARATE_LENGTH取一个点
                const res_length = path_res.path.length;
                if (res_length > CreepMotion.SEPARATE_LENGTH * 2) {
                    const end_reach = res_length - CreepMotion.SEPARATE_LENGTH - 1;
                    this.path = path_res.path.filter((p, i) => ((i + 1) % CreepMotion.SEPARATE_LENGTH === 0) && (i < end_reach));
                }
            }
            let stage_point = this.path_goal;
            if (this.path.length) {
                if (this.path_index < this.path.length) {  // 假设path检查是提前进行的，如果对path不满意，就清空它
                    stage_point = this.path[this.path_index];
                }

                // 如果creep与stage_point距离小于等于1，那么就更新path_index
                if (this.creep.__py__.distance(stage_point) <= 1) {
                    this.path_index++;
                    stage_point = this.path_index >= this.path.length ? this.path_goal : this.path[this.path_index];
                }
            }

            // std.info("name:", this.creep.name, "stage_point:", stage_point.x, stage_point.y);
            // this.view.rect(get.offset(stage_point, -.5, -.5), 1, 1, "#ff0000");

            return stage_point;
        }
    }

    clear() {
        /**
         * 清空所有信息
         * */
        this.path = [];
        this.path_goal = null;
        this.path_index = 0;
        this.lastPoint = {x: this.creep.x, y: this.creep.y};
    }

    handle() {
        // this.view.clear();
        if (this.creep.exists && this.creep.fatigue <= 0) {
            this.lastPoint = {x: this.creep.x, y: this.creep.y};  // 记录敌方creep上一帧的位置
        }
    }
}

class GameObject {
    static PREFIX_TABLE = {
        'Creep': "c",
        'StructureSpawn': "sp",
        'StructureExtension': "ext",
        'StructureRoad': "road",
        'StructureWall': "wall",
        'StructureRampart': "ramp",
        'StructureContainer': "box",
        'StructureTower': "twr",
        'Structure': "st",
        'ScoreCollector': "sc",
        'AreaEffect': "ae",
        'Source': "sr",
        'ConstructionSite': "site",
        'Flag': "flag",
        'Resource': "rs",
        'BodyPart': "bp",
        'GameObject': "go",
        'Object': "o",
    };
    static __PROTO__ = GameObjectProto;

    constructor(gobject) {
        this.__object__ = gobject;
        this.__object__.__py__ = this;
        this.__protoname__ = gobject.constructor.name;
        this.__name__ = gobject.name !== undefined ? gobject.name : `${GameObject.PREFIX_TABLE[this.__protoname__]}${gobject.id}`;

        // python
        this.py_name = this.name;
        this.__object__.name = this.name;
        this.__object__.py_name = this.py_name;
    }

    get name() {
        if (this.__name__ !== undefined) return this.__name__;
        this.__name__ = this.__object__.name !== undefined ? this.__object__.name : `${GameObject.PREFIX_TABLE[this.__protoname__]}${this.__object__.id}`;

        // python
        this.py_name = this.__name__;
        this.__object__.name = this.__name__;
        this.__object__.py_name = this.py_name;
    }

    set name(new_name) {
        if (!new_name) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.name`, `新名字不能为空. 本次操作无效.`);
            else std.warn(`${this.name}.name`, `new name can't be empty. This operation is invalid.`);
            return;
        }
        this.__name__ = new_name;
        this.__object__.name = new_name;
        this.__object__.py_name = new_name;
        this.py_name = new_name;
    }

    get id() { return this.__object__.id;}

    get _get_id() { return this.__object__.id;}

    get ticksToDecay() { return this.__object__.ticksToDecay;}

    get my() { return this.__object__.my;}

    get x() {
        return this.__object__.x;
    }

    get y() {
        return this.__object__.y;
    }

    get exists() {
        return this.__object__.exists;
    }

    distance(target) {
        if (!st.point(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.distance`, `目标:typeof ${typeof target}不是Point类型。忽略并返回65536`);
            else std.warn(`${this.name}.distance`, `target:typeof ${typeof target} is not Point type. Ignore and return 65536`);
            return 65536
        }
        return Math.max(Math.abs(this.x - target.x), Math.abs(this.y - target.y));
    }

    at(target) {
        return this.__object__.x === target.x && this.__object__.y === target.y;
    }

    nat(target) {
        return this.__object__.x !== target.x || this.__object__.y !== target.y;
    }

    offset(dx, dy) {
        return Point(this.x + dx, this.y + dy);
    }

    add(other) { return Point(this.x + other.x, this.y + other.y); }
    sub(other) { return Point(this.x - other.x, this.y - other.y); }
    neg() { return Point(-this.x, -this.y); }
    dot(other) { return this.x * other.x + this.y * other.y; }
    mul(other) { return Point(this.x * other, this.y * other); }
    crs(other) { return this.x * other.y - this.y * other.x; }

    closest(objs, filter_fn = undefined, notme = true) {
        if (len(objs) === 0) {
            return null;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.closest`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.closest`, 'No objects left after filtering.')
            return null;
        }
        if (notme) {
            objs = objs.filter(o => o !== this);
        }
        const ret = findClosestByRange(this, objs);
        if (ret === undefined) {
            return null;
        }
        return ret;
    }

    quickest(objs, filter_fn = undefined, options = undefined, notme = true) {
        if (len(objs) === 0) {
            return null;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.quickest`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.quickest`, 'No objects left after filtering.')
            return null;
        }
        if (notme) {
            objs = objs.filter(o => o !== this);
        }
        const ret = findClosestByPath(this, objs, options);
        if (ret === undefined) {
            return null;
        }
        return ret;
    }

    inrange(objs, range, filter_fn = null) {
        if (len(objs) === 0) {
            return objs;
        }
        if (filter_fn !== undefined && filter_fn !== null) {
            if (isinstance(filter_fn, tuple([list, tuple, set]))) {
                filter_fn = std.combo(filter_fn);
            }
            objs = filter(filter_fn, objs);
        }
        if (len(objs) === 0) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.inrange`, '过滤后没有剩余对象。');
            else std.warn(`${this.name}.inrange`, 'No objects left after filtering.')
            return null;
        }
        const tmp = [];
        for (const _obj of objs) {
            if (this.distance(_obj) <= range) {
                tmp.append(_obj);
            }
        }
        return tmp;
    }
    
    around(objects){
        /**
         * 这个函数获取unit 3范围内所有objects，然后按照health排序，记作around
         * 然后，获取unit 1范围内所有around，记作nears
         *
         * @return [nears, around]
         * */
        if (!objects){
            if (LANGUAGE === 'cn') std.warn("get.around", "参数 (!objects) 为True，请检查: typeof objects =", typeof objects);
            else std.warn("get.around", "(!objects) is true: typeof objects =", typeof objects);
            return [[], []];
        }
        const around = sorted(this.inrange(objects, 3), __kwargtrans__ ({key: (function (e) {
				return get.health (e, true);
			})}));
        const nears = this.inrange(around, 1);
        return [nears, around];
    }

    get _my_string_() {
        return this.__object__.my?"😄":(this.__object__.my === undefined?"😐":"😡");
    }

    toString() {
        return `(${this.name}${this._my_string_} ${this.x}◈${this.y})`;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof GameObject.__PROTO__;
    }
}

class Creep extends GameObject {
    static __PROTO__ = GameCreep;
    constructor(creep) {
        super(creep);
        this.motion = new CreepMotion(creep);
        this.info = new CreepInfo(creep);
        this._grade = this.info.grade;
        this._recipe = this.info.namedRecipe.recipe
        this._partsVector = this.info.partsVector;
        this._dynamic_next = math.irandom(100) + 100;
        this._flag = get.flag();
        this.__eda__ = new ListenSystem(this);
        this.__scs__ = new ScheduleSystem(this);

        // 为每个body内的元素添加py_metatype为其type
        for (const part of this.body) {
            part.py_metatype = part.type;
        }

        if (get.now === 1) {
            this.birthPoint = null;  // 这个应该是开局就有的
        } else {
            let sp = get.spawn(sp => sp.x === creep.x && sp.y === creep.y);
            this.birthPoint = sp ? Point(sp.x, sp.y) : null;
        }

        // 记录最近一个tick的操作内容
        this.actions = {
            move: null,     // Point
            attack: null,    // Creep | Structure
            // 如果同时近战和远程都攻击了，那么以该creep的优势能力的一项为基准
            heal: null,     // Creep
            harvest: null,  // Source
            build: null,    // Site
            fetch: null,    // Resource | Structure | Creep
            deposit: null,  // Creep | Structure | Point
        }
    }

    __update_info__() {
        if (!this._flag) this._dynamic_next = know.now + 999;
        else{
            this.info = new CreepInfo(this.__object__);
            this._recipe = this.info.namedRecipe.recipe
            this._partsVector = this.info.partsVector;
            this._grade = this.info.grade;
            this._dynamic_next = know.now + 40;
        }
    }

    get grade() {
        if (this._dynamic_next <= know.now) {
            this.__update_info__();
        }
        return this._grade;
    }

    get recipe() {
        if (this._dynamic_next <= know.now) {
            this.__update_info__();
        }
        return this._recipe;
    }

    get partsVector() {
        if (this._dynamic_next <= know.now) {
            this.__update_info__();
        }
        return this._partsVector;
    }

    get birthed() {
        if (this.birthPoint === null) return true;
        else return this.birthPoint.x !== this.__object__.x || this.birthPoint.y !== this.__object__.y; // 出生点变了, 说明出生了
    }
    get fatigue() {
        return this.__object__.fatigue;
    }

    get energy() { return this.__object__.store.getUsedCapacity(RESOURCE_ENERGY); }

    get energyMax() { return this.__object__.store.getCapacity(RESOURCE_ENERGY); }

    get hp() { return this.__object__.hits; }

    get hpMax() { return this.__object__.hitsMax; }
    get hits() {
        return this.__object__.hits;
    }

    get hitsMax() {
        return this.__object__.hitsMax;
    }


    get store() {
        return this.__object__.store;
    }

    get body() {
        return this.__object__.body;
    }

    handle() {
        /// 为每一个Creep提供更新, 外部调用. 应当在每个tick开始时调用
        this.motion.handle()
    }

    listenFor(event_name, callback, sources=undefined) {
        return this.__eda__.listenFor(event_name, callback, sources);
    }

    pushEvent(event_name, data) {
        return this.__eda__.pushEvent(event_name, data);
    }

    cancelListen(event_name){
        return this.__eda__.cancelListen(event_name);
    }

    schedule(delay, callback) {
        return this.__scs__.schedule(delay + know.now, callback);
    }

    move(to, options = undefined) {
        // 如果目标为 null 或 undefined，输出警告并返回 -1
        if (to === null || to === undefined) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.move`, `目标点是null或undefined。忽略此操作。`);
            else std.warn(`${this.name}.move`, `The target point is null or undefined. Ignore this operation.`);
            return ERR_INVALID_TARGET;
        }
        this.actions.move = to;


        // 初始化 options，确保其包含 costMatrix，否则使用默认值
        if (!options) options = {};
        else options = Object.assign({}, options);
        if (!options.costMatrix) options.costMatrix = DEFAULT_MOTION.costMatrix;

        if (options.plainCost !== undefined){
            if (LANGUAGE === "cn") std.warn(`${this.name}.move`, `由于使用了costMatrix，因此options.plainCost 已被禁用。`);
            else std.warn(`${this.name}.move`, `options.plainCost has been disabled since costMatrix is used.`);
        }

        if (options.swampCost !== undefined){
            if (options.swampCost !== 2){
                if (LANGUAGE === "cn") std.warn('put.move', `由于使用了costMatrix，因此options.swampCost 已被禁用(除了为2的情况)。`);
                else std.warn('put.move', `options.swampCost has been disabled since costMatrix is used(except for 2).`);
            } else {
                options.costMatrix = SWAMP_MOTION.costMatrix;
            }
        }


        let result = null;
        // print("to:", to);
        // 如果目标有 x 和 y 属性，计算下一个阶段点并移动单位
        if (to.x !== undefined && to.y !== undefined) {
            const dist = Math.max(Math.abs(this.x - to.x), Math.abs(this.y - to.y));
            if (dist === 0) return 0;
            else if (dist <= 2){
                const dirp = get.directionPoint(get.direction(this, to));
                const point = this.offset(dirp.x, dirp.y);
                // print("this:", this, "point:", point, "to:", to);
                if (DEFAULT_MOTION.costMatrix.get(point.x, point.y) < 254){  // 99 is creep's
                    return this.__object__.moveTo(point);
                }
            }
            const motion = this.motion;
            const stage_to = motion.seekNextStagePoint(to, options);
            const dist_to = this.distance(stage_to);
            if (!options.maxOps) options.maxOps = dist_to * dist_to * 2;
            result = this.__object__.moveTo(stage_to, options);
        }
        // 如果目标是整数，直接移动单位
        else if (Number.isInteger(to)) {
            result = this.__object__.move(to);
        }
        // 如果目标有 id 属性，但可能已被销毁，输出警告并设置结果为 ERR_INVALID_TARGET
        else if (to.id !== undefined) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.move`, `目标${std.__obj_str__(to)}可能已经被销毁。忽略本次操作。`);
            else std.warn(`${this.name}.move`, `Target ${std.__obj_str__(to)} may have been destroyed. Ignore this operation.`);
            result = ERR_INVALID_TARGET;
        }
        // 对于所有其他情况，目标不是有效的，输出警告并设置结果为 ERR_INVALID_TARGET
        else {
            if (LANGUAGE === "cn") std.warn(`${this.name}.move`, `目标${typeof to}不是有效的目标。忽略本次操作。`);
            else std.warn(`${this.name}.move`, `The target ${typeof to} is not a valid target. Ignore this operation.`);
            result = ERR_INVALID_TARGET;
        }

        return result;
    }

    escape(target, options = undefined) {
        if (!st.point(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.escape`, `目标${std.__obj_str__(target)}不是有效的目标。忽略本次操作。`);
            else std.warn(`${this.name}.escape`, `The target ${std.__obj_str__(target)} is not a valid target. Ignore this operation.`);
            return ERR_INVALID_TARGET;
        }
        const dist = this.distance(target);
        if (dist <= 3) {
            const around = this.inrange(know.battles.enemies, dist + 1);
            target = Point.center(...around);
        }
        if (target.x === this.x && target.y === this.y) {
            // 随机选择[1, 8]
            const random = math.irandrange(1, 8);
            return this.move(random);
        }
        const dx = this.x - target.x;
        const dy = this.y - target.y;
        let to = Point(math.clamp(this.x + dx * 2, 0, 100), math.clamp(this.y + dy * 2, 0, 100));
        to = get.space(to.x, to.y);
        if (this.x === 48 && this.y === 50) print(this, target, to);
        return this.move(to, options);
    }

    __judge__(dm, dr, edm, edr) {
        /// 这个函数判断是否应该靠近目标
        if (dm === 0 && edm > 0) return false;
        if (dm === 0 && edm === 0) {
            if (dr === 0) return false;
            return !(dr > 0 && edr >= dr - 1);

        } else if (dm > 0 && edm > 0) {
            if (dr === 0) return true;  // 你不能远程，你死都得上
            if (dm - edm > 1) return true;
            const melee_bonus = dm * 3 / (dm * 3 + dr);
            const enemy_melee_bonus = edm * 3 / (edm * 3 + edr);
            return melee_bonus >= enemy_melee_bonus - 0.1;
        }
        return true;
    }

    attack(target, move = true) {
        /**
         * 指挥一个单位攻击指定的目标，可选地包含移动逻辑。
         *
         * @param {st.hitable} target - 被攻击的目标对象。
         * @param {Boolean|Object} [move=true] - 是否包含移动逻辑(false表明不移动)，可传递移动选项。
         * @returns {Number) - 一个代表操作结果的常量。
         */
        if (!st.hitable(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.attack`, `目标${std.__obj_str__(target)}不可被攻击。`);
            else std.warn(`${this.name}.attack`, `Target ${std.__obj_str__(target)} cannot be attacked.`);
        }
        this.actions.attack = target;
        const dist = this.distance(target);
        if (dist > 3) {
            if (move) return this.move(target, (move === true ? undefined : move));
            return ERR_NOT_IN_RANGE;
        }
        const melee = st.melee(this.__object__);
        const ranged = st.ranged(this.__object__);
        const dynamicMelees = this.info.dynamicPartsVector.melees;
        const dynamicRanges = this.info.dynamicPartsVector.ranges;
        const enemyDynamicMelees = (target instanceof Creep) ? target.info.dynamicPartsVector.melees : 1;
        const enemyDynamicRanges = (target instanceof Creep) ? target.info.dynamicPartsVector.ranges : 0;

        let melee_result = ERR_NO_BODYPART;
        let ranged_result = ERR_NO_BODYPART;
        if (dist === 1) {
            if (melee) melee_result = this.__object__.attack(target.__object__);
            if (ranged) ranged_result = this.__object__.rangedMassAttack();
            if (move && this.__judge__(dynamicMelees, dynamicRanges, enemyDynamicMelees, enemyDynamicRanges)) {
                this.move(target, (move === true ? undefined : move));
            }
        } else {
            if (ranged) {
                ranged_result = this.__object__.rangedAttack(target.__object__);
                if (move && dist <= 2) {
                    this.escape(target, (move === true ? undefined : move));
                }
            } else if (move && this.__judge__(dynamicMelees, dynamicRanges, enemyDynamicMelees, enemyDynamicRanges)) {
                this.move(target, (move === true ? undefined : move));
                melee_result = ERR_NOT_IN_RANGE;
            }
        }

        if (dynamicMelees > dynamicRanges) return melee_result;
        else if (dynamicMelees < dynamicRanges) return ranged_result;
        else return ERR_NO_BODYPART;
    }

    heal(target, move = true) {
        /**
         * 执行单位的治疗操作。
         *
         * @param {Creep} target - 治疗的目标单位。
         * @param {boolean} [move=true] - 是否允许单位移动以接近目标。
         * @returns {Object} - 一个包含治疗结果的对象，可能的属性有 `heal`、`ranged` 或 `move`。
         */
        // 检查目标是否有效
        if (!st.creep(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.heal`, `目标${std.__obj_str__(target)}不可被治疗。`);
            else std.warn(`${this.name}.heal`, `Target ${std.__obj_str__(target)} cannot be healed.`);
        }
        this.actions.heal = target;

        // 初始化结果对象
        let result = null;

        // 计算单位与目标之间的距离
        const dist = this.distance(target);

        // 如果距离小于等于1，尝试近战治疗
        if (dist <= 1) result = this.__object__.heal(target.__object__);
        // 如果距离在2到3之间，尝试远程治疗
        else if (dist <= 3) result = this.__object__.rangedHeal(target.__object__);
        // 如果距离大于3且允许移动，移动单位接近目标
        else if (move) {
            this.move(target, (move === true ? undefined : move));
            result = ERR_NOT_IN_RANGE;
        }

        // 返回治疗结果
        return result;
    }

    __induce_resource_type__(target) {
        if (!target){
            if (LANGUAGE === "cn") std.warn(`${this.name}.__induce_resource_type__`, `目标为:${target}。默认为能量。`);
            else std.warn(`${this.name}.__induce_resource_type__`, `The target is:${target}. Default to energy.`);
            return RESOURCE_ENERGY;
        }
        if (target instanceof Source) return RESOURCE_ENERGY;
        if (target instanceof Resource) return target.resourceType;
        if (target instanceof ScoreCollector) return target.resourceType;
        if (!target.store) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.__induce_resource_type__`, `目标${std.__obj_str__(target)}没有store的能力。默认为能量类型。`);
            else std.warn(`${this.name}.__induce_resource_type__`, `The target ${std.__obj_str__(target)} has no store ability. Default to energy type.`);
            return RESOURCE_ENERGY;
        }
        const store = target.store;
        const cap_energy = store.getUsedCapacity(RESOURCE_ENERGY);
        const cap_score = store.getUsedCapacity(RESOURCE_SCORE);
        const cap_score_x = store.getUsedCapacity(RESOURCE_SCORE_X);
        const cap_score_y = store.getUsedCapacity(RESOURCE_SCORE_Y);
        const cap_score_z = store.getUsedCapacity(RESOURCE_SCORE_Z);
        // print(this.name, cap_energy, cap_score, cap_score_x, cap_score_y, cap_score_z)
        if(cap_energy > 0) return RESOURCE_ENERGY;
        else if(cap_score > 0) return RESOURCE_SCORE;
        else if(cap_score_x > 0) return RESOURCE_SCORE_X;
        else if(cap_score_y > 0) return RESOURCE_SCORE_Y;
        else if(cap_score_z > 0) return RESOURCE_SCORE_Z;
        if (LANGUAGE === "cn") std.warn(`${this.name}.__induce_resource_type__`, `目标未检测到任何资源类型。默认返回能量类型。`);
        else std.warn(`${this.name}.__induce_resource_type__`, `No resource type detected for the target. Default to energy type.`);
        return RESOURCE_ENERGY;

    }

    fetch(target, resource_type = null, amount = null, move = true) {
        /**
         * fetch - 从指定目标获取资源。
         *
         * 该函数用于从目标获取资源。它可以处理移动到目标、捡起资源、从结构中提取资源或从另一个单位转移资源。
         *
         * @param {Resource|Creep|Structure} target - 资源的来源。
         * @param {String} [resource_type=RESOURCE_ENERGY] - 要获取的资源类型。
         * @param {number} [amount=null] - 要获取的资源数量。
         * @param {boolean|Object} [move=true] - 是否移动到目标。如果为 `true`，则使用默认移动选项。
         * @returns {number} 操作结果的错误代码。
         */
        // 检查目标是否有效
        if (!st.storable(target) && !(target instanceof Resource)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.fetch`, `目标${std.__obj_str__(target)}没有store的能力。`);
            else std.warn(`${this.name}.fetch`, `Target ${std.__obj_str__(target)} has no store ability.`);
            return ERR_INVALID_TARGET;
        }
        this.actions.fetch = target;

        if (resource_type === null || resource_type === undefined) resource_type = this.__induce_resource_type__(target);

        // 如果需要移动且单位与目标距离大于1，则移动到目标
        if (move && this.distance(target) > 1) {
            return this.move(target, (move === true ? undefined : move));
        }

        // 如果目标是资源对象且为能量，则尝试捡起
        if ((target instanceof Resource) && target.resourceType === RESOURCE_ENERGY) return this.__object__.pickup(target.__object__);

        // 根据目标类型和是否指定了数量，执行转移或提取操作
        if (amount) {
            if (target instanceof Creep) return target.__object__.transfer(this.__object__, resource_type, amount);
            if (target instanceof Structure) return this.__object__.withdraw(target.__object__, resource_type, amount);
        } else {
            if (target instanceof Creep) return target.__object__.transfer(this.__object__, resource_type);
            if (target instanceof Structure) return this.__object__.withdraw(target.__object__, resource_type);
        }

        // 如果以上所有操作均失败，则返回目标无效错误
        return ERR_INVALID_TARGET;
    }

    deposit(target, resource_type = null, amount = null, move = true) {
        /**
         * deposit - 将资源存储到指定目标或丢弃资源。
         *
         * 该函数用于将 `this` 的资源存储到 `target`，如果 `target` 为 `null`，则丢弃资源。
         * 它可以处理移动到目标、转移资源或丢弃资源。
         *
         * @param {Structure|Creep|null} target - 资源的目标。可以是结构、单位或 `null` 以丢弃资源。
         * @param {String} [resource_type=RESOURCE_ENERGY] - 要存储的资源类型。如果为 `null`，则默认为能量。
         * @param {number} [amount=null] - 要存储的资源数量。如果为 `null`，则存储所有该类型的资源。
         * @param {boolean|Object} [move=true] - 是否移动到目标。如果为 `true`，则使用默认移动选项。
         * @returns {number} 操作结果的错误代码。
         */
        this.actions.deposit = target;

        // 如果目标为 null，则丢弃资源
        if (target === null) {
            this.actions.deposit = {x: this.__object__.x, y: this.__object__.y};
            if (resource_type === null || resource_type === undefined) resource_type = this.__induce_resource_type__(this);
            return this.__object__.drop(resource_type, amount);
        }

        if (resource_type === null || resource_type === undefined) resource_type = this.__induce_resource_type__(target);

        if (!st.point(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.deposit`, `目标${std.__obj_str__(target)}不是一个有效的目标。`);
            else std.warn(`${this.name}.deposit`, `The target ${std.__obj_str__(target)} is not a valid target.`);
            return ERR_INVALID_TARGET;
        }

        // 如果需要移动且单位与目标距离大于1，则移动到目标
        if (move && this.distance(target) > 1) {
            return this.move(target, (move === true ? undefined : move));
        }

        // 根据目标是否可存储和是否指定了数量，执行转移或丢弃操作
        if (amount) {
            if (st.storable(target)) return this.__object__.transfer(target.__object__, resource_type, amount);

        } else {
            if (st.storable(target)) return this.__object__.transfer(target.__object__, resource_type);
        }

        this.actions.deposit = {x: this.__object__.x, y: this.__object__.y};
        return this.__object__.drop(resource_type);
    }

    build(site, move = true) {
        /**
         * build - 执行建造操作。
         *
         * 该函数用于在指定位置建造结构。如果单位与建造位置的距离大于3，它会先移动到该位置。
         *
         * @param {ConstructionSite} site - 建造的目标位置。
         * @param {boolean|Object} [move=true] - 是否移动到建造位置。如果为 `true`，则使用默认移动选项。
         * @returns {number} 操作结果的错误代码。
         */

        // 检查建造位置是否有效
        if (!st.site(site)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.build`, `目标${std.__obj_str__(site)}不是有效的建造位置。`);
            else std.warn(`${this.name}.build`, `The target ${std.__obj_str__(site)} is not a valid construction site.`);
        }
        this.actions.build = site;

        // 如果需要移动且单位与建造位置距离大于3，则移动到该位置
        if (move && this.distance(site) > 3) {
            return this.move(site, (move === true ? undefined : move));
        }

        // 执行建造操作
        return this.__object__.build(site.__object__);
    }

    harvest(target, move = true) {
        /**
         * harvest - 执行采集操作。
         *
         * 该函数用于从资源点采集资源。如果单位与资源点的距离大于1，它会先移动到该位置。
         * 如果目标不是资源点，会发出警告并忽略该命令。
         *
         * @param {Source} target - 采集的目标，必须是资源点。
         * @param {boolean|Object} [move=true] - 是否移动到资源点。如果为 `true`，则使用默认移动选项。
         * @returns {number} 操作结果的错误代码。
         */
        // 检查目标是否有效
        if (!st.source(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.harvest`, `目标${std.__obj_str__(target)}不是有效的资源点。`);
            else std.warn(`${this.name}.harvest`, `The target ${std.__obj_str__(target)} is not a valid source.`);
        }
        this.actions.harvest = target;

        // 如果需要移动且单位与资源点距离大于1，则移动到该位置
        if (move && this.distance(this, target) > 1) {
            return this.move(target, (move === true ? undefined : move));
        }

        // 执行采集操作
        return this.__object__.harvest(target.__object__);
    }

    intermit(target, resource_type = null, options = null) {
        /**
         * intermit - 促进单位以“断断续续”的方式搬运资源。
         *
         * 该函数通过让 `this` 在移动时放下和重新捡起资源，从而有效地将资源搬运到 `target`。
         * 这种方法特别适用于移动能力有限但携带能力高的单位，因为它可以减少移动时的疲劳积累。
         *
         * @param {Structure|Creep|null} target - 资源的目标位置。可以是结构、单位或 `null` 以放下资源。
         * @param {String} [resource_type=RESOURCE_ENERGY] - 要搬运的资源类型。
         * @param {Object|null} [options=null] - 移动选项。如果为 `null`，则使用默认选项。
         * @returns {number} 操作结果的错误代码。
         */
        // 设置资源类型为能量，如果未指定
        if (!st.point(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.intermit`, `目标${std.__obj_str__(target)}不是有效的位置。`);
            else std.warn(`${this.name}.intermit`, `The target ${std.__obj_str__(target)} is not a valid point.`);
            return ERR_INVALID_TARGET;
        }

        if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;

        let result = null;

        // 获取单位最后掉落的资源
        const drops = get.resources(function (obj) {
            return obj.resourceType === resource_type && this._last_idrop && this._last_idrop.x === obj.x && this._last_idrop.y === obj.y;
        }.bind(this));

        // 获取单位的剩余存储空间
        const free = this.store.getFreeCapacity(resource_type);

        // 如果有掉落的资源且单位有剩余存储空间，尝试捡起这些资源
        if (len(drops) && free > 0) {
            return this.fetch(drops[0], resource_type, free, false);
        }

        // 如果单位没有携带任何该类型的资源，完成操作
        const current = this.store.getUsedCapacity(resource_type);
        if (current === 0) {
            return DONE;
        }

        // 检查目标是否有效
        if (!st.storable(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.intermit`, `目标${std.__obj_str__(target)}不是有效的存储位置。`);
            else std.warn(`${this.name}.intermit`, `The target ${std.__obj_str__(target)} is not a valid storage location.`);
        }

        // 计算单位与目标之间的距离
        const dist = this.distance(target);

        // 如果距离小于等于1，直接将资源存储到目标
        if (dist <= 1) {
            result = this.deposit(target, resource_type, null, (options ? options : true));
            if (result !== OK) {
                return result;
            }
            return DONE;
        }

        // 如果距离大于1，先放下资源，然后移动到目标
        else {
            result = this.deposit(null, resource_type, null, (options ? options : true));
            if (this._last_idrop) {
                this._last_idrop.x = this.x;
                this._last_idrop.y = this.y;
            } else {
                this._last_idrop = Point(this.x, this.y);
            }
            if (result !== OK) {
                return result;
            }
            return this.move(target, options);
        }
    }

    follow(target, distance, options = null) {
        /**
         * follow - 管理单位跟随目标的行为。
         *
         * 该函数根据 `target` 是否属于我方，控制 `this` 的移动行为。如果 `target` 属于我方，
         * `this` 会尝试保持在 `stop_dist` 和 `ignore_dist` 之间的距离。如果 `target` 不属于我方，
         * `this` 会在距离过近或 `target` 有疲劳时尝试逃逸。
         *
         * 如果目标是 Creep，那么评估的距离是 `this` 到目标的下一个预测点的距离。
         *
         * @param {Creep|Structure} target - 被跟随的目标。
         * @param {number|tuple} distance - 与目标保持的距离。可以是一个数或一个元组 `[stop_dist, ignore_dist]`。
         * @param {Object|null} [options=null] - 移动选项。如果为 `null`，则使用默认选项。
         * @returns {number} 操作结果的错误代码。
         */
        // 检查目标是否有效
        if (!st.point(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.follow`, `目标${std.__obj_str__(target)}不是有效的位置。`);
            else std.warn(`${this.name}.follow`, `The target ${std.__obj_str__(target)} is not a valid position.`);
            return ERR_INVALID_TARGET;
        }

        let result = null;
        const target_is_creep = st.creep(target);

        // 处理 distance 参数，确保它是一个元组 [stop_dist, ignore_dist]
        const __left0__ = (Array.isArray(distance) ? distance : [distance, 0]);
        const stop_dist = __left0__[0];
        const ignore_dist = __left0__[1];

        // 计算单位与目标之间的距离
        const pre_target = (target instanceof Creep) ? target.motion.seekNext() : target;
        const dist = this.distance(pre_target);

        // 分支逻辑基于 target 是否属于我方
        if (target.my) {
            // 目标属于我方
            if (target_is_creep && (ignore_dist >= dist && dist >= stop_dist)) {
                // 如果在 stop_dist 和 ignore_dist 之间，则尝试移动 target 向 this
                result = target.move(this, options);
                if (result !== OK) {
                    return result;
                }
            }
            // 否则，移动 this 向 target
            return this.move(target, options);
        } else {
            // 目标不属于我方
            print(this, "dist:", dist, "stop_dist:", stop_dist, "fatigue:", target.fatigue);
            if (dist < stop_dist || (dist === stop_dist && target.fatigue !== undefined && target.fatigue <= 0)) {
                // 如果距离过近或 target 没有疲劳，则尝试逃逸
                return this.escape(target, options);
            } else if (dist > stop_dist) {
                // 否则，移动 this 向 target
                return this.move(target, options);
            }
            return OK;  // 否则不需要移动
        }
    }

    carry(src, dst, resource_type = null, options = null, intermit = false) {
        /**
         * carry - 管理单位在源点和目标之间搬运资源。
         *
         * 该函数处理 `this` 从 `src` 到 `dst` 的资源搬运。它根据 `intermit` 参数决定是采用常规搬运还是“断断续续”搬运。
         * 在常规搬运中，`this` 直接从 `src` 搬运资源到 `dst`。在断断续续搬运中，`this` 会先放下资源，然后移动到 `dst`，
         * 以减少移动时的疲劳积累，特别适用于移动能力有限但携带能力高的单位。
         *
         * @param {Structure|Creep} src - 资源的来源位置。
         * @param {Structure|Creep} dst - 资源的目标位置。
         * @param {String} [resource_type=RESOURCE_ENERGY] - 要搬运的资源类型。
         * @param {Object|null} [options=null] - 移动选项。如果为 `null`，则使用默认选项。
         * @param {boolean} [intermit=false] - 是否采用“断断续续”的搬运方式。但是注意，如果你的moves>=carrys，在平地上仍然会使用非intermit的方式。
         * @returns {number} 操作结果的错误代码。如果搬运完成，返回 `DONE`。
         */
        // 设置资源类型为能量，如果未指定
        if (!st.point(dst)){
            if (LANGUAGE === "cn") std.warn(`${this.name}.carry`, `参数dst:${std.__obj_str__(dst)}不是有效的位置。`);
            else std.warn(`${this.name}.carry`, `The parameter dst:${std.__obj_str__(dst)} is not a valid position.`);
            return ERR_INVALID_TARGET;
        }
        if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;

        // 分支到常规搬运或断断续续搬运，基于 intermit 参数
        if (!intermit) {
            // 常规搬运逻辑
            if (this.store.getUsedCapacity(resource_type) === 0) {
                // 单位没有该类型的资源
                if (!st.storable(src)) {
                    if (LANGUAGE === "cn") std.warn(`${this.name}.carry`, `参数src:${std.__obj_str__(src)}不是有效的存储位置。`);
                    else std.warn(`${this.name}.carry`, `The parameter src:${std.__obj_str__(src)} is not a valid storage position.`);
                }
                if (this.distance(src) <= 1) {
                    const _ = this.fetch(src, resource_type, null, false);
                    return this.move(dst, options);
                }
                return this.fetch(src, resource_type, null, (options ? options : true));
            }
            if (this.distance(dst) <= 1) {
                this.deposit(dst, resource_type, null, false);
                if (!st.storable(dst)) {
                    if (LANGUAGE === "cn") std.warn(`${this.name}.carry`, `参数dst:${std.__obj_str__(dst)}不是有效的存储位置。`);
                    else std.warn(`${this.name}.carry`, `The parameter dst:${std.__obj_str__(dst)} is not a valid storage position.`);
                }
                return this.move(src, options);
            }
            const result = this.deposit(dst, resource_type, null, (options ? options : true));
            if (result < 0) return result;

            // 检查一下目标是否有足够的裕量
            const eng0 = this.store.getUsedCapacity(resource_type);
            const cap0 = dst.store.getFreeCapacity(resource_type);
            if (cap0 < eng0) return 0;
            else return DONE;
        } else {
            // 断断续续搬运逻辑
            let result = null;
            const drops = get.resources(function (obj) {
                return obj.resourceType === resource_type && this._last_idrop && this._last_idrop.x === obj.x && this._last_idrop.y === obj.y;
            }.bind(this));
            const free = this.store.getFreeCapacity(resource_type);

            // print("drops:", drops.length, "free:", free);
            if (drops.length > 0 && free > 0) {
                return this.fetch(drops[0], resource_type, free, false);
            }

            const current = this.store.getUsedCapacity(resource_type);
            // print("current:", current);
            if (current === 0) {
                if (!st.storable(src)) {
                    if (LANGUAGE === "cn") std.warn(`${this.name}.carry`, `参数src:${std.__obj_str__(src)}不是有效的存储位置。`);
                    else std.warn(`${this.name}.carry`, `The parameter src:${std.__obj_str__(src)} is not a valid storage position.`);
                }
                return this.fetch(src, resource_type, null, (options ? options : true));
            }

            const dist = this.distance(dst);
            // print("dist:", dist);
            if (dist <= 1) {
                result = this.deposit(dst, resource_type, null, (options ? options : true));
                if (result !== OK) {
                    return result;
                }

                // 检查一下目标是否有足够的裕量
                const eng0 = this.store.getUsedCapacity(resource_type);
                const cap0 = dst.store.getFreeCapacity(resource_type);
                if (cap0 < eng0) return 0;
                else {
                    if (st.storable(src)) this.move(src, options);
                    return DONE;
                }
            } else {
                if (this.fatigue > 0) return ERR_TIRED;
                // 判断move能力是否足够
                if ((this.partsVector.moves - this.partsVector.carries - this.partsVector.others) < 0 || get.COST_ARRAY2D[this.y][this.x] >= 10){
                    result = this.deposit(null, resource_type, null, (options ? options : true));
                    if (this._last_idrop) {
                        this._last_idrop.x = this.x;
                        this._last_idrop.y = this.y;
                    } else {
                        this._last_idrop = Point(this.x, this.y);
                    }
                    if (result !== OK) {
                        return result;
                    }
                } else {
                    this._last_idrop = undefined;
                }
                return this.move(dst, options);
            }
        }
    }

    patrol(points, options = null) {
        /**
         * patrol - 指令 creep 按顺序循环访问 points 中的所有点。
         *
         * @param {Point[]} points - 要访问的点的数组。
         * @param {Object} [options=null] - 移动选项。
         * @returns {number} 返回移动 move 结果
         */
            // 生成巡逻点的哈希值
        let hash = 0;
        for (let i = 0; i < points.length; i++) {
            hash |= points[i].x;
            hash <<= 1;
            hash |= points[i].y;
            hash <<= 1;
            hash &= 0x7FFFFFFF;
        }

        // 如果哈希值发生变化，重新初始化巡逻路径和索引
        if (this.patrol_hash === undefined || this.patrol_hash !== hash) {
            // print("patrol hash changed:", this.patrol_hash, hash);
            this.patrol_hash = hash;
            this.patrol_path = points;
            this.patrol_index = 0;

            // 遍历所有点，找到最近的点
            let min_dist = 999;
            for (let i = 0; i < points.length; i++) {
                const dist = this.distance(points[i]);
                if (dist < min_dist) {
                    min_dist = dist;
                    this.patrol_index = i;
                }
            }
        }

        // print("patrol:", this.patrol_index, this.patrol_path.length);
        // 获取当前目标点
        const pt = this.patrol_path[this.patrol_index];
        const dist = this.distance(pt);

        // 如果距离小于等于2，则切换到下一个点
        if (dist <= 2) {
            this.patrol_index = (this.patrol_index + 1) % this.patrol_path.length;
        }

        // 移动到当前目标点
        return this.move(pt, options);
    }

    pull(target, options = null) {
        if (!(target instanceof Creep)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.pull`, `目标${std.__obj_str__(target)}不是Creep类型。`);
            else std.warn(`${this.name}.pull`, `Target ${std.__obj_str__(target)} is not a Creep type.`);
        }
        if (this.distance(target) > 1) {
            this.move(target, options);
            return ERR_NOT_IN_RANGE;
        }
        target.move(this);
        return this.__object__.pull(target.__object__);
    }

    push(target, options = null) {
        if (!st.movable(target)) {
            if (LANGUAGE === "cn") std.warn(`${this.name}.push`, `目标${std.__obj_str__(target)}不可被推动。`);
            else std.warn(`${this.name}.push`, `Target ${std.__obj_str__(target)} cannot be pushed.`);
        }

        target.__object__.pull(this.__object__);
        if (this.distance(target) > 1) return this.move(target, options);
        return this.move(target);
    }

    __umin__(array, callback, log=false) {
        /**
         * @param {Array} array -
         * @param {Function} callback - 用于计算的回调函数。
         * @param {boolean} [log=false] - 是否打印结果。
         **/
        // 计算callback(item) => number 的结果
        const results = [];
        for (let i = 0; i < array.length; i++) { results.push(callback(array[i], i)); }
        if (log) {
            const prints = [];
            for (let i = 0; i < array.length; i++) prints.push({name: array[i].name, x: array[i].x, y: array[i].y, result: results[i]});
            std.log(`${this.name}.__umin__`, prints);
        }
        // 返回最小值的对应的原数组的值
        let _min_val = 9999;
        let _min_index = 0;
        for (let i = 0; i < results.length; i++) {
            if (results[i] < _min_val) {
                _min_val = results[i];
                _min_index = i;
            }
        }
        if (log) std.log(`${this.name}.__umin__`, `min_index:${_min_index}, min_val:${_min_val}`);
        return array[_min_index];
    }

    static LIMIT_SEARCH = {maxOps:9};
    // 通用逻辑
    autoAttack(view_range = 50, wall=true) {
        /**
         * autoAttack - 管理单位的自动攻击逻辑。
         *
         * 该函数根据单位周围的敌人类型和优先级，决定单位的近战和远程攻击目标。
         * 它首先选择近战目标，然后选择远程目标，确保单位在近战和远程范围内有效攻击敌人。
         * 该函数还处理目标锁定，确保单位持续跟踪最接近的高优先级敌人。
         *
         * @param {number} [view_range=50] - 单位的视野范围。
         * @param {boolean} [wall=true] - 是否允许攻击墙壁。
         * @returns {Array} 返回一个包含两个元素的数组：
         *   - `locked`：锁定的目标，用于持续跟踪。
         *   - `target`：该tick被攻击的目标，如果附近被攻击的目标，返回null。
         */
        let target = null;
        let _melee_target = null;  // 仅用于记录近战目标
        let locked = null;
        let extreme = false;

        // 获取周围不同类型的敌人
        let [nears_battle, around_battle] = this.around(know.battles.enemies, true);
        let [nears_civilian, around_civilian,] = this.around(know.civilian.enemies);
        let [nears_spawns, around_spawns,] = this.around(get.spawns(st.enemy));
        let [nears_extensions, around_extensions,] = this.around(get.extensions(st.enemy));
        let [nears_towers, around_towers,] = this.around(get.towers(st.enemy));
        let [nears_ramparts, around_ramparts,] = this.around(get.ramparts(st.enemy));
        let [nears_walls, around_walls,] = wall?this.around(get.walls()):[[],[]];

        // 选择近战攻击目标
        if (len(nears_battle)) {
            // 按照dynamicAttackPower + dynamicHealPower来排序
            target = this.__umin__(nears_battle, u => u.hp * (u.info.dynamicAttackPower + u.info.dynamicHealPower)?0.9:1);
            locked = nears_battle[0];
        } else if (len(nears_civilian)) target = nears_civilian[0];
        else if (len(nears_extensions)) target = nears_extensions[0];
        else if (len(nears_towers)) target = nears_towers[0];
        else if (len(nears_spawns)) target = nears_spawns[0];
        else if (len(nears_ramparts)) target = nears_ramparts[0];
        else if (len(nears_walls)) target = nears_walls[0];

        // 执行近战攻击
        if (target) this.attack(target, false);
        _melee_target = target;
        target = null;

        // 选择远程攻击目标
        if (!locked && len(around_battle)) {
            target = around_battle[0];

            if (this.info.dynamicPartsVector.melees){
                // 避免重复计算
                const _calculated = {};
                // 避免极端情况
                const _ignores = {};
                locked = this.__umin__(around_battle, u => {
                    const hp_ratio = u.hits / u.hitsMax;

                    const tmp_dist = this.distance(u);
                    const dist_ratio = 1;

                    const path = get.path(this, u, Creep.LIMIT_SEARCH);
                    if (path.length <= 0) return tmp_dist * dist_ratio + hp_ratio;
                    // 判断目标点附近1范围和距离2的敌人数量
                    const p = Point(path[0].x, path[0].y);
                    // 检查是否已经计算过
                    let score = _calculated[`${p.x},${p.y}`];
                    if (score !== undefined) return score + hp_ratio;

                    const melees2 = [];
                    const melees1 = [];
                    for (const e of around_battle) {
                        if (p.distance(e) <= 1) melees1.push(e);
                        if (p.distance(e) <= 2) melees2.push(e);
                    }
                    // melees的melees的dynamicMelees总和越大，评分越高
                    let melees1_score = 0;
                    let melees2_score = 0;
                    for (const melee of melees1) melees1_score += melee.info.dynamicPartsVector.melees;
                    if (melees1_score > this.info.dynamicPartsVector.melees * 1.8) _ignores[`${u.x},${u.y}`] = true;
                    for (const melee of melees2) melees2_score += melee.fatigue<=0?melee.info.dynamicPartsVector.melees:0;
                    score = tmp_dist * dist_ratio + melees1_score + melees2_score * 0.25;
                    print(`p:${p} melees1:${melees1.length} melees2:${melees2.length} score:${score} ${score + hp_ratio}`);
                    _calculated[`${p.x},${p.y}`] = score;
                    return score + hp_ratio;
                });
                // 避免极端情况
                if (_ignores[`${locked.x},${locked.y}`]) {
                    locked = null;
                    extreme = true;
                }
            } else if (!locked) locked = around_battle[0];
        } else if (len(around_civilian)) target = around_civilian[0];
        else if (len(around_extensions)) target = around_extensions[0];
        else if (len(around_towers)) target = around_towers[0];
        else if (len(around_spawns)) target = around_spawns[0];
        else if (len(around_ramparts)) target = around_ramparts[0];
        else if (len(around_walls)) target = around_walls[0];

        // 锁定目标
        if (locked === null && len(know.battles.enemies) > 0 && !extreme) {
            const tmp_es = this.inrange(know.battles.enemies, view_range);
            if (len(tmp_es) > 0) locked = this.closest(tmp_es);
        }

        // 执行远程攻击
        if (target) this.attack(target, false);

        // 返回锁定目标和近战目标
        return [locked, _melee_target ? _melee_target : target];
    }
    autoHeal(view_range = 50, lock_healer=true) {
        /**
         * autoHeal - 管理单位的自动治疗逻辑。
         *
         * 该函数根据单位周围的友方单位的健康状况，决定单位的治疗目标。
         * 它首先检查近战范围内的战斗单位，然后是周围范围内的战斗单位，
         * 接着是近战范围内的平民单位，最后是周围范围内的平民单位。
         * 如果没有立即的治疗目标，它会锁定视野范围内的最接近的战斗单位，
         * 以确保持续的治疗支持。
         *
         * @param {number} [view_range=50] - 单位的视野范围。
         * @returns {Array} 返回一个包含两个元素的数组：
         *   - `locked`：锁定的目标，用于持续跟踪。
         *   - `target`：当前治疗的目标。
         */
        let target = null;
        let locked = null;

        // 获取周围不同类型的友方单位
        let [nears_battles, around_battles] = this.around(know.battles.friends);
        let [nears_civilians, around_civilians] = this.around(know.civilian.friends);

        // 选择治疗目标
        if (len(nears_battles) && nears_battles[0].hits < nears_battles[0].hitsMax) {
            if (lock_healer || nears_battles[0].partsVector.heals <= 0) locked = nears_battles[0];
            target = nears_battles[0];
        } else if (len(around_battles) && around_battles[0].hits < around_battles[0].hitsMax) {
            target = around_battles[0];
            if (lock_healer || nears_battles[0].partsVector.heals <= 0) locked = around_battles[0];
        } else if (len(nears_civilians) && nears_civilians[0].hits < nears_civilians[0].hitsMax) {
            target = nears_civilians[0];
        } else if (len(around_civilians) && around_civilians[0].hits < around_civilians[0].hitsMax) {
            target = around_civilians[0];
        }

        // 锁定目标
        if (!locked) {
            const tmp_es = this.inrange(know.battles.friends, view_range, u => (this !== u && (lock_healer || u.partsVector.heals <= 0)));
            if (len(tmp_es) > 0) locked = this.closest(tmp_es);
        }

        // 执行治疗
        if (target) this.heal(target, false);

        // 返回锁定目标和治疗目标
        return [locked, target];
    }

    details() { return this.info.details(); }

    toString() {
        if (!this.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            const capacityMax = this.store.getCapacity();
            const capacityString = ` 📦${this.store.getUsedCapacity()}/${capacityMax}`;
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y} ⭐️${this.grade} ❤️${this.hits}/${this.hitsMax}${capacityMax ? capacityString : ""})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof Creep.__PROTO__;
    }

}

class Structure extends GameObject {
    static __PROTO__ = GameStructure;
    constructor(object) {
        super(object);
        this.hitsMax = object.hitsMax;
    }

    get hits() {
        return this.__object__.hits;
    }

    get hp() { return this.__object__.hits; }

    get hpMax() { return this.__object__.hitsMax; }

    toString() {
        if (!this.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y} ❤️${this.hits}/${this.hitsMax})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof Structure.__PROTO__;
    }
}

class Resource extends GameObject {
    static __PROTO__ = GameResource;
    constructor(object) {
        super(object);
        this.resourceType = object.resourceType;
    }

    get amount() {
        return this.__object__.amount;
    }

    get energy() { return this.amount; }

    toString() {
        if (!this.exists) {
            return `(${this.name}☠️)`;
        } else {
            return `(${this.name} ${this.x}◈${this.y} ${this.resourceType} 📦${this.amount})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof Resource.__PROTO__;
    }
}

class Source extends GameObject {
    static __PROTO__ = GameSource;
    constructor(object) {
        super(object);
        this.energyCapacity = object.energyCapacity;
    }

    get energy() {
        return this.__object__.energy;
    }

    get energyMax() { return this.energyCapacity; }

    toString() {
        if (!this.exists) {
            return `(${this.name}☠️)`;
        } else {
            return `(${this.name} ${this.x}◈${this.y} 📦${this.energy}/${this.energyCapacity})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof Source.__PROTO__;
    }
}

class StructureContainer extends Structure {
    static __PROTO__ = GameStructureContainer;
    constructor(object) {
        super(object);
        this.cost = 200;
        this.capacity = this.energyMax;
    }

    get store() {
        return this.__object__.store;
    }

    get energy() { return this.__object__.store.getUsedCapacity(RESOURCE_ENERGY); }

    get energyMax() { return this.__object__.store.getCapacity(RESOURCE_ENERGY); }

    toString() {
        if (!this.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            let decay_string = "";
            if (this.ticksToDecay !== undefined) decay_string = ` ⏳${this.ticksToDecay}t`;
            const energyMax = this.energyMax;
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y} ❤️${this.hits}/${this.hitsMax} 📦${this.energy}/${energyMax>=0?energyMax:"?"}${decay_string})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureContainer.__PROTO__;
    }
}

class StructureExtension extends StructureContainer {
    static __PROTO__ = GameStructureExtension;
    constructor(object) {
        super(object);
        this.cost = 200;
        this.capacity = this.energyMax;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureExtension.__PROTO__;
    }
}

class ScoreCollector extends GameObject {
    static __PROTO__ = GameScoreCollector;
    constructor(object) {
        super(object);
        this.capacity = object.scoreTotal;
        this.scoreTotal = object.scoreTotal;
        this.resourceType = RESOURCE_SCORE;
    }

    get my() { return this.__object__.my; }

    get score() { return this.__object__.score; }

    get scoreMax() { return this.__object__.scoreTotal; }

    get scoreX() { return this.resourceType === RESOURCE_SCORE_X?this.__object__.score:0; }

    get scoreXMax() { return this.resourceType === RESOURCE_SCORE_X?this.__object__.scoreTotal:0; }

    get scoreY() { return this.resourceType === RESOURCE_SCORE_Y?this.__object__.score:0; }

    get scoreYMax() { return this.resourceType === RESOURCE_SCORE_Y?this.__object__.scoreTotal:0; }

    get scoreZ() { return this.resourceType === RESOURCE_SCORE_Z?this.__object__.score:0; }

    get scoreZMax() { return this.resourceType === RESOURCE_SCORE_Z?this.__object__.scoreTotal:0; }

    toString() {
        if (!this.exists) {
            return `(${this.name}☠️)`;
        } else {
            return `(${this.name} ${this.x}◈${this.y} ${this.resourceType} 📦${this.score}/${this.scoreTotal})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof ScoreCollector.__PROTO__;
    }
}

class AreaEffect extends GameObject {
    static __PROTO__ = GameAreaEffect;
    constructor(object) {
        super(object);
        this.effect = object.effect;
    }

    toString() {
        if (!this.exists) {
            return `(${this.name}☠️)`;
        } else {
            let decay_string = "";
            if (this.ticksToDecay !== undefined) decay_string = ` ⏳${this.ticksToDecay}t`;
            return `(${this.name} ${this.x}◈${this.y} ${this.effect}${decay_string})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof AreaEffect.__PROTO__;
    }
}


class StructureWall extends Structure {
    static __PROTO__ = GameStructureWall;
    constructor(object) {
        super(object);
        this.cost = 100;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureWall.__PROTO__;
    }
}

class StructureRampart extends Structure {
    static __PROTO__ = GameStructureRampart;
    constructor(object) {
        super(object);
        this.cost = 200;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureRampart.__PROTO__;
    }
}

class StructureRoad extends Structure {
    static __PROTO__ = GameStructureRoad;
    constructor(object) {
        super(object);
        this.cost = get.terrain(this.x, this.y) === TERRAIN_SWAMP ? 50 : 10;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureRoad.__PROTO__;
    }
}

class StructureTower extends StructureContainer {
    static __PROTO__ = GameStructureTower;
    constructor(object) {
        super(object);
        this.cost = 1250;
        this.capacity = this.energyMax;
    }

    get cooldown() {
        return this.__object__.cooldown;
    }

    measure(target) {
        /// 计算对目标的伤害或是治疗量
        let dist = this.distance(target);
        // Attack effectiveness	150 hits at range ≤5 to 37 hits at range ≥20
        // Heal effectiveness	100 hits at range ≤5 to 25 hits at range ≥20
        if (dist > 50) return 0;

        dist = math.clamp(dist, 5, 20);

        if (target.my) {
            return math.lerp(100, 25, (dist - 5) / 15);
        } else {
            return math.lerp(150, 37, (dist - 5) / 15);
        }
    }

    attack(target) {
        return this.__object__.attack(target.__object__);
    }

    heal(target) {
        return this.__object__.heal(target.__object__);
    }

    toString() {
        if (!this.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            const energyMax = this.energyMax;
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y} ❤️${this.hits}/${this.hitsMax} 📦${this.energy}/${energyMax>=0?energyMax:"?"} 🔄${this.cooldown})t`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureTower.__PROTO__;
    }

}

class SpawnRecipeManager {
    constructor() {
        this.data = {};
    }

    // 添加数组到对应长度的组中
    add(arr_or_has_recipe) {
        const obj = arr_or_has_recipe.recipe?arr_or_has_recipe: {recipe: arr_or_has_recipe};
        const length = obj.recipe.length;
        if (!this.data[length]) {
            this.data[length] = []; // 每个长度对应一个数组
        }
        this.data[length].push(obj); // 插入的对象包含 recipe 属性
    }

    // 弹出指定长度和索引的数组
    pop(length, index) {
        if (!this.data[length] || index < 0 || index >= this.data[length].length) {
            return undefined;
        }
        const popped = this.data[length].splice(index, 1);
        if (this.data[length].length === 0) {
            delete this.data[length];
        }
        return popped && popped[0]; // 返回弹出的对象
    }

    // 获取指定长度和索引的数组
    get(length, index) {
        if (!this.data[length] || index < 0 || index >= this.data[length].length) {
            return undefined;
        }
        return this.data[length][index]; // 返回对象
    }

    /**
     * 在给定的数组中查找匹配的索引。
     *
     * @param {Array} arr - 需要查找的数组。
     * @returns {Number} - 匹配的索引，如果未找到则返回 -1。
     */
    index(arr) {
        // 获取输入数组的长度
        const length = arr.length;

        // 从数据对象中获取与数组长度匹配的可能项
        const possibleMatches = this.data[length];

        // 如果没有可能的匹配项，返回 -1
        if (!possibleMatches) return -1;

        // 初始化一个对象，用于记录在当前迭代中需要跳过的匹配项
        const skips = {};

        // 初始化结果变量
        let result_index = -1;
        let result_flag = false;

        // 遍历输入数组的每个元素
        for (let i = 0; i < length; i++) {
            result_flag = false; // 重置结果标志

            // 遍历每个可能的匹配项
            for (let j = 0; j < possibleMatches.length; j++) {
                // 如果匹配项在当前迭代中被跳过，继续下一个
                if (skips[j]) continue;

                // 检查当前元素是否匹配
                if (possibleMatches[j].recipe[i] !== arr[i]) {
                    // 如果不匹配，标记为跳过
                    skips[j] = true;
                } else {
                    // 如果匹配，更新结果索引并设置标志
                    result_index = j;
                    result_flag = true;
                }
            }

            // 如果在任何位置没有找到匹配项，提前退出
            if (!result_flag) break;
        }

        // 如果找到了匹配项，返回其索引；否则返回 -1
        return result_flag ? result_index : -1;
    }

    // 移除指定的数组
    remove(arr) {
        const length = arr.length;
        const idx = this.index(arr);
        if (idx !== -1) {
            this.pop(length, idx);
        }
    }

    // 获取当前数据结构
    getData() {
        return this.data;
    }
}

class StructureSpawn extends StructureContainer {
    static __PROTO__ = GameStructureSpawn;
    constructor(object) {
        super(object);
        this.cost = 3000;
        this.capacity = 1000;
        this.__current__ = {
            creep:null,
            name: null,
            recipe: [],
            directions: [],
            nextTick: 0,
        }
        this.__records__ = new SpawnRecipeManager();
    }

    get spawning() {
        return this.__object__.spawning;
    }

    get needTime() {
        if (this.__object__.spawning) return this.__object__.spawning.needTime;
        else return 0;
    }

    get remainingTime() {
        if (this.__object__.spawning) return this.__object__.spawning.remainingTime;
        else return -1;
    }

    get creep() {
        if (this.__object__.spawning) return this.__object__.creep;
        else return null;
    }

    cancel() {
        if (this.__object__.spawning) {
            return this.__object__.spawning.cancel();
        }
        return ERR_NOT_FOUND;
    }

    __compareStringArrays__(arr1, arr2) {
        // 如果长度不相等，立即返回false
        if (arr1.length !== arr2.length) {
            return false;
        }
        // 如果两个数组长度均为0，立即返回true
        if (arr1.length === 0) {
            return true;
        }
        // 如果两个数组长度均为1，直接比较唯一元素
        if (arr1.length === 1) {
            return arr1[0] === arr2[0];
        }
        // 遍历数组，逐项比较
        for (let i = 0; i < arr1.length; i++) {
            if (arr1[i] !== arr2[i]) {
                return false;
            }
        }
        // 如果所有项都相同，返回true
        return true;
    }

    create(recipe, creep_name = null, optimise = true, direction = null) {
        /**
         * create - 从 spawn 使用指定的配方创建一个 creep。
         *
         * @param {String[]} recipe - 用于创建 creep 的身体部位数组。
         * @param {string} [creep_name=null] - 可选的 creep 名称。
         * @param {boolean} [optimise=true] - 是否优化配方以提高功能完存率。
         * @param {number} [direction=null] - 可选的方向参数，用于指定 creep 诞生时的方向。
         *               如果诞生时指定的方向被占据，会自动尝试相邻的方向。如果仍然无法诞生，该参数会被忽略。
         * @returns {Creep|number} 返回新创建的 creep 或剩余的孵化时间(>0)，或错误代码(<0)。
         */

        // 检查第一帧(remainingTime is 0)重复同一recipe的情况
        if (this.remainingTime === -1 && get.now < this.__current__.nextTick) {
            return this.__current__.recipe.length * CREEP_SPAWN_TIME - 1;
        }
        // 空闲 或 与当前配方不同
        const equal = this.__compareStringArrays__(this.__current__.recipe, recipe);
        if (this.remainingTime > 0 && equal) return this.__object__.spawning.remainingTime - 1;
        if (this.remainingTime <= 0 || !equal) {
            // 检查__records__中是否有相同的配方
            const index = this.__records__.index(recipe);
            if (index !== -1) {

                const result = this.__records__.get(recipe.length, index);

                this.__records__.pop(recipe.length, index);
                // console.log("creep:", result.creep);
                return new Creep(result.creep);
            }
        }

        if (this.remainingTime <= 0) {  // 此处已经排除了先前生产过的情况
            // 优化配方
            const raw_recipe = recipe;
            if (optimise) recipe = PartsVector.partsOptimise(recipe);

            // 尝试孵化 creep
            const result = this.__object__.spawnCreep(recipe);
            if (result.error !== undefined) return result.error;
            // console.log("result:", result);
            if (creep_name) result.object.name = creep_name;
            this.__current__ = {
                creep: result.object,
                name: creep_name,
                recipe: raw_recipe,
                directions: direction !== null?get.expandDirection(direction):[TOP, TOP_RIGHT, RIGHT, BOTTOM_RIGHT, BOTTOM, BOTTOM_LEFT, LEFT, TOP_LEFT],
                nextTick: get.now + 3,
                costTime: recipe.length * CREEP_SPAWN_TIME,
            }
            this.__records__.add(this.__current__);
            // console.log("set Directions:", this.__current__.directions);
            this.__object__.setDirections(this.__current__.directions);
            return recipe.length * CREEP_SPAWN_TIME - 1;  // 返回剩余孵化时间
        }

        return ERR_BUSY;
    }

    static CLOCKS = [
        "🕛", "🕐", "🕑", "🕒", "🕓", "🕔",
        "🕕", "🕖", "🕗", "🕘", "🕙", "🕚", "🕛"
    ];

    static PERIOD = 1.0 / 13;

    toString() {
        if (!this.__object__.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            let time_string = "☕"
            if (this.remainingTime > 0) {
                const ratio = 1 - this.remainingTime / this.__current__.costTime;
                time_string = `${StructureSpawn.CLOCKS[Math.floor(ratio / StructureSpawn.PERIOD)]}${this.remainingTime}t`;
            }
            const energyMax = this.energyMax;
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y} ❤️${this.hits}/${this.hitsMax} 📦${this.energy}/${energyMax>=0?energyMax:"?"} ${time_string})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof StructureSpawn.__PROTO__;
    }
}

class ConstructionSite extends GameObject {
    static __PROTO__ = GameConstructionSite;
    constructor(object) {
        super(object);
        this.type = undefined;  // 用于记录建筑类型

    }

    get py_metatype() { return this.type; }

    static plan(x, y, building_type, single = false) {
        /**
         * site - 在指定位置创建一个建筑工地。
         *
         * 该函数用于在游戏地图上的指定位置创建一个建筑工地。如果 `single` 参数为 `true`，
         * 它会确保没有其他相同类型的工地已经存在于该位置。如果尝试创建重复的工地，
         * 则返回 `ERR_NAME_EXISTS`。否则，它会尝试创建工地并返回结果。
         *
         * @param {number} x - 建筑工地的 x 坐标。
         * @param {number} y - 建筑工地的 y 坐标。
         * @param {object} building_type - 要创建的建筑类型。
         * @param {boolean} [single=false] - 是否确保工地唯一。如果为 `true`，则不能创建重复的工地。
         * @returns {ConstructionSite|number} 返回创建的工地对象或错误代码。
         */
        if (!building_type || !building_type.__PROTO__) return ERR_INVALID_ARGS;
        // 如果 single 为 true，检查是否已经存在相同类型的工地
        if (single) {
            const exists = get.site(function (ste) {
                return ste.x === x && ste.y === y && ste.type === building_type;
            });
            if (exists) {
                return ERR_NAME_EXISTS;
            }
        }

        // 尝试创建建筑工地
        const result = createConstructionSite(x, y, building_type.__PROTO__);
        if (result.object !== undefined) {
            // 如果成功，返回工地对象
            const site = new ConstructionSite(result.object);
            site.type = building_type;
            return site;
        }
        // 如果失败，返回错误代码
        return result.error;
    }

    get progressTotal() { return this.__object__.progressTotal; }

    get progress() { return this.__object__.progress; }

    get structure() { return this.__object__.structure; }

    get percentage() { return this.progress / this.progressTotal; }


    remove() { return this.__object__.remove(); }

    toString() {
        if (!this.__object__.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            return `(${this.name}:${this.type.name}${this._my_string_} ${this.x}◈${this.y} 🔧${this.progress}/${this.progressTotal})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof ConstructionSite.__PROTO__;
    }

}

class Flag extends GameObject {
    static __PROTO__ = GameFlag;
    constructor(object) {
        super(object);
    }

    toString() {
        if (!this.exists) {
            return `(${this.name}${this._my_string_}☠️)`;
        } else {
            return `(${this.name}${this._my_string_} ${this.x}◈${this.y})`;
        }
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof Flag.__PROTO__;
    }
}

class BodyPart extends GameObject {
    static __PROTO__ = GameBodyPart;
    constructor(object) {
        super(object);
        this.type = object.type;

        //
        this.py_metatype = this.type;
    }

    toString() {
        let decay_string = "";
        if (this.ticksToDecay !== undefined) decay_string = ` ⏳${this.ticksToDecay}t`;
        return `(${this.name}:${this.type} ${this.x}◈${this.y}${decay_string})`;
    }

    static [Symbol.hasInstance](obj) {
        return obj && obj.__object__ instanceof BodyPart.__PROTO__;
    }
}

const PYEMBED_TABLE = {
    'Creep': Creep,
    'ConstructionSite': ConstructionSite,
    'StructureSpawn': StructureSpawn,
    'StructureExtension': StructureExtension,
    'StructureRoad': StructureRoad,
    'StructureWall': StructureWall,
    'StructureRampart': StructureRampart,
    'StructureContainer': StructureContainer,
    'StructureTower': StructureTower,
    'Structure': Structure,
    'Source': Source,
    'Resource': Resource,
    'BodyPart': BodyPart,
    'Flag': Flag,
    'ScoreCollector': ScoreCollector,
    'AreaEffect': AreaEffect,
    'GameObject': GameObject,
};

const std = {
    __obj_str__: function (obj) {
        /**
         * 这个函数尝试对一个obj进行字符串化
         * */
        if (obj === undefined) return 'undefined';
        if (obj === null) return 'null';
        try {
            return obj.toString();
        } catch (e) {
            if (obj.name !== undefined) return `$${obj.name}`;
            if (obj.id !== undefined) return `$${obj.id}`;
            return `$${typeof obj}`;
        }
    },
    __expand_composite_as_list__: function (filter_composite) {
        var tmp = [];
        if (isinstance(filter_composite, tuple([tuple, list]))) {
            for (var each of filter_composite)
                tmp.extend(std.__expand_composite_as_list__(each));
            return tmp;
        }
        if (isinstance(filter_composite, set))
            return std.__expand_composite_as_list__(py_next(py_iter(filter_composite)));
        return [filter_composite];
    },
    __generate_combo_js_eval__: function (filter_composite, fid) {
        if (isinstance(filter_composite, list)) {
            return ('(' + ' || '.join((function () {
                var __accu0__ = [];
                for (var each of filter_composite) {
                    __accu0__.append(std.__generate_combo_js_eval__(each, fid));
                }
                return py_iter(__accu0__);
            })())) + ')';
        }
        if (isinstance(filter_composite, tuple)) {
            return ('(' + ' && '.join((function () {
                var __accu0__ = [];
                for (var each of filter_composite) {
                    __accu0__.append(std.__generate_combo_js_eval__(each, fid));
                }
                return py_iter(__accu0__);
            })())) + ')';
        }
        if (isinstance(filter_composite, set))
            return (' !(' + std.__generate_combo_js_eval__(py_next(py_iter(filter_composite)), fid)) + ')';
        return ('%' + fid [filter_composite]) + '%';

    },
    __expand_memory__: dict({}),
    combo: function (filter_composite) {
        const memory_key = str(filter_composite);
        const fn = std.__expand_memory__.py_get(memory_key, null);
        if (fn) return fn;
        const fn_ids = dict({});
        const id_fns = dict({});
        const _ = std.__expand_composite_as_list__(filter_composite);
        for (var [f_id, each_fn] of enumerate(set(_))) {
            f_id++;
            fn_ids [each_fn] = f_id;
            id_fns [f_id] = each_fn;
        }
        var _eval_str = std.__generate_combo_js_eval__(filter_composite, fn_ids);
        var _inner = function (obj) {
            var eval_str = _eval_str + '';
            for (var [fn_id, func] of id_fns.py_items()) {
                eval_str = eval_str.replace('%' + fn_id + '%', String(!!func(obj)));
                // pass;
            }
            return eval(eval_str);
        };
        std.__expand_memory__ [memory_key] = _inner;
        return _inner;
    },
    log: function (caller_name, ...args) {
        console.log("[" + caller_name, " [33m" + "Log" + "[0m" + "]:", ...args)
    },
    info: function (caller_name, ...args) {
        console.log("[" + caller_name, " [42m" + "Info" + "[0m" + "]:", ...args)
    },
    warn: function (caller_name, ...args) {
        console.log("[" + caller_name, " [43m" + "Warn" + "[0m" + "]:", ...args)
    },
    error: function (caller_name, ...args) {
        throw new Error("[" + caller_name + " [41m" + "Warn" + "[0m" + "]: " + args.join(" "))
    },
    ansi_color: function (r, g = undefined, b = undefined, bg = false) {
        /**
         * ansi_color - 生成 ANSI 颜色代码字符串。
         *
         * 该函数根据提供的 RGB 值或十六进制颜色字符串生成 ANSI 颜色代码。
         * 它支持生成前景颜色和背景颜色代码。如果输入无效或 RGB 值超出范围，
         * 它会报告错误并返回 ANSI 重置代码。
         *
         * @param {string|number} r - 十六进制颜色字符串（以 `#` 开头）或红色分量（0-255）。
         * @param {number} [g=undefined] - 如果 `r` 是数字，则为绿色分量（0-255）。如果 `r` 是字符串，则为背景颜色标志。
         * @param {number} [b=undefined] - 如果 `r` 是数字，则为蓝色分量（0-255）。
         * @param {boolean} [bg=false] - 是否生成背景颜色代码。
         * @returns {string} ANSI 颜色代码字符串或 ANSI 重置代码 `JS_RESET`，如果输入无效。
         */
        let red, green, blue;

        // 如果第一个参数是字符串，假设是十六进制颜色字符串
        if (typeof r === 'string' && r.startsWith('#')) {
            red = parseInt(r.slice(1, 3), 16);
            green = parseInt(r.slice(3, 5), 16);
            blue = parseInt(r.slice(5, 7), 16);
            bg = g;
        }
        // 否则，假设是三个 RGB 值
        else if (typeof r === 'number' && typeof g === 'number' && typeof b === 'number') {
            red = r;
            green = g;
            blue = b;
        } else {
            if (LANGUAGE === 'cn') {
                std.error('ansi_color', '无效的颜色输入', r, g, b);
            } else {
                std.error('ansi_color', 'Invalid color input', r, g, b);
            }
            return JS_RESET;
        }

        // 检查 RGB 值是否在有效范围内
        if (red < 0 || red > 255 || green < 0 || green > 255 || blue < 0 || blue > 255) {
            if (LANGUAGE === 'cn') {
                std.error('ansi_color', '无效的 RGB 值', r, g, b);
            } else {
                std.error('ansi_color', 'Invalid RGB value', r, g, b);
            }
            return JS_RESET;
        }

        // 构造 ANSI 颜色代码字符串
        return bg ? `\x1b[48;2;${red};${green};${blue}m` : `\x1b[38;2;${red};${green};${blue}m`;
    },

    WRAP_END_TABLE: {
        '(': ')',
        '[': ']',
        '{': '}',
        '<': '>',
        '"': '"',
        "'": "'"
    },
    wrap_bracket: function (str, bracket, color) {
        /**
         * @description: 在字符串两端添加指定的括号和颜色
         * @param {string} str - 要添加括号的字符串
         * @param {string} bracket - 括号类型，例如'(', '[', '{', '<', '"', "'"中的一种
         * @param {string} color - 颜色代码，例如'\x1b[32m'表示绿色
         * */
        const end_bracket = std.WRAP_END_TABLE [bracket];
        if (end_bracket === undefined)
            if (LANGUAGE === 'cn') std.error('wrap_bracket', '括号类型错误:', bracket);
            else std.error('wrap_bracket', 'Bracket type error:', bracket);

        return `${color}${bracket}${JS_RESET}${str}${color}${end_bracket}${JS_RESET}`;
    },

    print_map: function (rect_) {
        /**
         * print_map - 打印游戏地图的指定矩形区域。
         *
         * 该函数遍历地图的指定矩形区域，从 `rect_.top` 到 `rect_.bottom` 和从 `rect_.left` 到 `rect_.right`。
         * 对于每个位置 (x, y)，它从 `COST_MATRIX` 中获取值，并根据值的不同使用不同的字符表示：
         * - 数字（<10）：直接打印数字。
         * - 10：打印 `#`。
         * - 64：打印 `C`。
         * - 其他情况：打印 `X`。
         * 每一行构建完毕后，使用 `console.log` 打印该行。
         *
         * @param {object} rect_ - 定义要打印的矩形区域的对象，包含 `top`、`bottom`、`left` 和 `right` 属性。
         * @returns {void} 无返回值。输出直接打印到控制台。
         */
        /// 这个函数用于打印地图中的一个矩形区域，<10的部分使用数字填充, 10使用#填充, 64使用C填充, 其他使用X填充。
        for (var y = rect_.top; y <= rect_.bottom; y++) {
            if (y < 0 || y >= 100) continue;  // 跳过无效的 y 坐标
            var line = '';
            for (var x = rect_.left; x <= rect_.right; x++) {
                if (x < 0 || x >= 100) continue;  // 跳过无效的 x 坐标
                var value = get.COST_ARRAY2D[y][x];  // 获取当前位置的值
                if (value < 10) line += value.toString() + " ";  // 数字 <10
                else if (value === 10) line += '# ';  // 数字 10
                else if (value <= 200) line += 'Z ';  // 数字 50
                else if (value === 254) line += 'C ';  // 数字 254
                else line += 'X ';  // 数字 255
            }
            console.log(line);  // 打印构建的行
        }
    },
    broadcast: function (arr, func = (x, y, val) => val, start = {x: 0, y: 0}, end = undefined) {
        /**
         * broadcast - 对二维数组的指定区域中的每个元素应用函数，并将结果存储回原数组。
         *
         * 该函数遍历二维数组 `arr` 的指定区域，从 `start` 到 `end`，对每个元素应用 `func` 函数。
         * `func` 函数接受当前元素的 `x` 和 `y` 坐标以及当前值 `val`，并返回新值，该值将存储回 `arr[y][x]`。
         * 如果 `end` 未指定，则 `end` 会被设置为 `arr` 的形状。
         *
         * @param {number[][]} arr - 要操作的二维数组。
         * @param {function(x:number, y:number, val:number): number} [func=(x, y, val)=>val] - 要应用到每个元素的函数。
         * @param {object} [start={x:0, y:0}] - 指定起始坐标 `{x:0, y:0}`。
         * @param {object} [end=undefined] - 指定结束坐标。如果为 `undefined`，则设置为 `arr` 的形状。
         * @returns {void} 无返回值。原数组 `arr` 被修改。
         */
        /// 允许对一个2d数组中指定区域的元素进行逐个操作，并将返回值存储在原数组中。
        if (end === undefined) end = get.shape(arr);
        for (let x = start.x; x < end.x; x++) {
            for (let y = start.y; y < end.y; y++) {
                arr[y][x] = func(x, y, arr[y][x]);
            }
        }
    },

    __cpu_limit__: Math.floor(ARENA_CPU_TIME_LIMIT / 1000),
    __cpu_first_limit__: Math.floor(ARENA_CPU_TIME_LIMIT_FIRST_TICK / 1000000),
    show_welcome: function () {
        // 获取当前竞技场的信息
        const info = get.arena();
        const arena_name = info.py_name.__getslice__(0, 2, 1).lower();

        // 根据不同的竞技场名称，调整背景颜色
        var bg_color = '\x1b[48;2;45;45;45m';
        if (arena_name === 'sp') bg_color = '\x1b[48;2;45;64;39m';
        else if (arena_name === 'ca') bg_color = '\x1b[48;2;38;50;80m';
        else if (arena_name === 'co') bg_color = '\x1b[48;2;70;46;31m';

        // 根据语言设置，打印欢迎信息
        if (LANGUAGE === 'cn') {
            console.log(bg_color + '---------- 欢迎来到' + '\x1b[1m\x1b[38;2;205;85;85mScreeps \x1b[38;2;65;105;225mArena\x1b[22m' + '\x1b[37m' + '! ----------' + '\x1b[0m');
            console.log('工具库版本: ' + __VERSION__ + ', Python版本: ' + __PYTHON_VERSION__, '\n');
        } else {
            console.log(bg_color + '---------- Welcome to ' + '\x1b[1m\x1b[38;2;205;85;85mScreeps \x1b[38;2;65;105;225mArena\x1b[22m' + '\x1b[37m' + '! ----------' + '\x1b[0m');
            console.log('Package Version: ' + __VERSION__ + ', Python Version: ' + __PYTHON_VERSION__, '\n');
        }
    },
    show_usage: function () {
        const game_tick = get.ticks();

        // 根据语言设置，打印硬件使用情况的标题
        if (LANGUAGE === 'cn') console.log('[性能使用情况]');
        else console.log('[Hardware Usage]');

        // 显示CPU使用情况
        if (game_tick <= 1) {
            var v = round(get.cpu() / 1000000);
            var color = (v < 800 ? (v < 600 ? '\x1b[32m' : '\x1b[33m') : '\x1b[31m');
            console.log('cpu: ' + (color + v.toString()) + '\x1b[0m' + ' ms / {} ms'.format(std.__cpu_first_limit__));
        } else {
            var v = round(get.cpu() / 1000);
            var color = (v < 75000 ? (v < 50000 ? '\x1b[32m' : '\x1b[33m') : '\x1b[31m');
            console.log('cpu: ' + (color + v.toString()) + '\x1b[0m' + ' us / {} us'.format(std.__cpu_limit__));
        }

        // 显示内存使用情况
        const heap = get.heap();
        var v = round(heap.used_heap_size / 1024);
        var v_limit = round(heap.heap_size_limit / 1024);
        var color = (v < v_limit * 0.6 ? '\x1b[32m' : (v < v_limit * 0.8 ? '\x1b[33m' : '\x1b[31m'));
        if (LANGUAGE === 'cn') {
            console.log('内存: ' + (color + v.toString()) + '\x1b[0m' + ' KB / ' + v_limit + ' KB');
        } else {
            console.log('heap: ' + (color + v.toString()) + '\x1b[0m' + ' KB / ' + v_limit + ' KB');
        }
    },
}
